#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterable, Sequence, no_type_check
import copy
import numpy as np

from ezdxf import colors
from ezdxf.math import Vec2, BoundingBox2d, Matrix44
from ezdxf.path import Command

from .type_hints import Color
from .backend import BackendInterface, BkPath2d, BkPoints2d, ImageData
from .config import Configuration, LineweightPolicy
from .properties import BackendProperties
from . import layout, recorder


__all__ = ["PlotterBackend"]

SEMICOLON = ord(";")
PRELUDE = b"%0B;IN;BP;"
EPILOG = b"PU;PA0,0;"
FLATTEN_MAX = 10  # plot units
MM_TO_PLU = 40  # 40 plu = 1mm
DEFAULT_PEN = 0
WHITE = colors.RGB(255, 255, 255)
BLACK = colors.RGB(0, 0, 0)
MAX_FLATTEN = 10

# comparing Command.<attrib> to ints is very slow
CMD_MOVE_TO = int(Command.MOVE_TO)
CMD_LINE_TO = int(Command.LINE_TO)
CMD_CURVE3_TO = int(Command.CURVE3_TO)
CMD_CURVE4_TO = int(Command.CURVE4_TO)


class PlotterBackend(recorder.Recorder):
    """The :class:`PlotterBackend` creates HPGL/2 plot files for output on raster
    plotters. This backend does not need any additional packages.  This backend support
    content cropping at page margins.

    The plot files are tested by the plot file viewer `ViewCompanion Standard`_
    but not on real hardware - please use with care and give feedback.

    .. _ViewCompanion Standard: http://www.softwarecompanions.com/

    """

    def __init__(self) -> None:
        super().__init__()

    def get_bytes(
        self,
        page: layout.Page,
        *,
        settings: layout.Settings = layout.Settings(),
        render_box: BoundingBox2d | None = None,
        curves=True,
        decimal_places: int = 1,
        base=64,
    ) -> bytes:
        """Returns the HPGL/2 data as bytes.

        Args:
            page: page definition, see :class:`~ezdxf.addons.drawing.layout.Page`
            settings: layout settings, see :class:`~ezdxf.addons.drawing.layout.Settings`
            render_box: set explicit region to render, default is content bounding box
            curves: use Bèzier curves for HPGL/2 output
            decimal_places: HPGL/2 output precision, less decimal places creates smaller
                files but for the price of imprecise curves (text)
            base: base for polyline encoding, 32 for 7 bit encoding or 64 for 8 bit encoding

        """
        top_origin = False
        settings = copy.copy(settings)
        # This player changes the original recordings!
        player = self.player()
        if render_box is None:
            render_box = player.bbox()

        # the page origin (0, 0) is in the bottom-left corner.
        output_layout = layout.Layout(render_box, flip_y=False)
        page = output_layout.get_final_page(page, settings)
        if page.width == 0 or page.height == 0:
            return b""  # empty page
        # DXF coordinates are mapped to integer coordinates (plu) in the first
        # quadrant: 40 plu = 1mm
        settings.output_coordinate_space = (
            max(page.width_in_mm, page.height_in_mm) * MM_TO_PLU
        )
        # transform content to the output coordinates space:
        m = output_layout.get_placement_matrix(
            page, settings=settings, top_origin=top_origin
        )
        player.transform(m)
        if settings.crop_at_margins:
            p1, p2 = page.get_margin_rect(top_origin=top_origin)  # in mm
            # scale factor to map page coordinates to output space coordinates:
            output_scale = settings.page_output_scale_factor(page)
            max_sagitta = 0.1 * MM_TO_PLU  # curve approximation 0.1 mm
            # crop content inplace by the margin rect:
            player.crop_rect(p1 * output_scale, p2 * output_scale, max_sagitta)
        backend = _RenderBackend(
            page,
            settings=settings,
            curves=curves,
            decimal_places=decimal_places,
            base=base,
        )
        player.replay(backend)
        return backend.get_bytes()

    def compatible(
        self, page: layout.Page, settings: layout.Settings = layout.Settings()
    ) -> bytes:
        """Returns the HPGL/2 data as 7-bit encoded bytes curves as approximated
        polylines and coordinates are rounded to integer values.
        Has often the smallest file size and should be compatible to all output devices
        but has a low quality text rendering.
        """
        return self.get_bytes(
            page, settings=settings, curves=False, decimal_places=0, base=32
        )

    def low_quality(
        self, page: layout.Page, settings: layout.Settings = layout.Settings()
    ) -> bytes:
        """Returns the HPGL/2 data as 8-bit encoded bytes, curves as Bézier
        curves and coordinates are rounded to integer values.
        Has a smaller file size than normal quality and the output device must support
        8-bit encoding and Bèzier curves.
        """
        return self.get_bytes(
            page, settings=settings, curves=True, decimal_places=0, base=64
        )

    def normal_quality(
        self, page: layout.Page, settings: layout.Settings = layout.Settings()
    ) -> bytes:
        """Returns the HPGL/2 data as 8-bit encoded bytes, curves as Bézier
        curves and coordinates are floats rounded to one decimal place.
        Has a smaller file size than high quality and the output device must support
        8-bit encoding, Bèzier curves and fractional coordinates.
        """
        return self.get_bytes(
            page, settings=settings, curves=True, decimal_places=1, base=64
        )

    def high_quality(
        self, page: layout.Page, settings: layout.Settings = layout.Settings()
    ) -> bytes:
        """Returns the HPGL/2 data as 8-bit encoded bytes and all curves as Bézier
        curves and coordinates are floats rounded to two decimal places.
        Has the largest file size and the output device must support 8-bit encoding,
        Bèzier curves and fractional coordinates.
        """
        return self.get_bytes(
            page, settings=settings, curves=True, decimal_places=2, base=64
        )


class PenTable:
    def __init__(self, max_pens: int = 64) -> None:
        self.pens: dict[int, colors.RGB] = dict()
        self.max_pens = int(max_pens)

    def __contains__(self, index: int) -> bool:
        return index in self.pens

    def __getitem__(self, index: int) -> colors.RGB:
        return self.pens[index]

    def add_pen(self, index: int, color: colors.RGB):
        self.pens[index] = color

    def to_bytes(self) -> bytes:
        command: list[bytes] = [f"NP{self.max_pens-1};".encode()]
        pens: list[tuple[int, colors.RGB]] = [
            (index, rgb) for index, rgb in self.pens.items()
        ]
        pens.sort()
        for index, rgb in pens:
            command.append(make_pc(index, rgb))
        return b"".join(command)


def make_pc(pen: int, rgb: colors.RGB) -> bytes:
    # pen color
    return f"PC{pen},{rgb.r},{rgb.g},{rgb.b};".encode()


class _RenderBackend(BackendInterface):
    """Creates the HPGL/2 output.

    This backend requires some preliminary work, record the frontend output via the
    Recorder backend to accomplish the following requirements:

    - Move content in the first quadrant of the coordinate system.
    - The output coordinates are integer values, scale the content appropriately:
        - 1 plot unit (plu) = 0.025mm
        - 40 plu = 1mm
        - 1016 plu = 1 inch
        - 3.39 plu = 1 dot @300 dpi
    - Replay the recorded output on this backend.

    """

    def __init__(
        self,
        page: layout.Page,
        *,
        settings: layout.Settings,
        curves=True,
        decimal_places: int = 2,
        base: int = 64,
    ) -> None:
        self.settings = settings
        self.curves = curves
        self.factional_bits = round(decimal_places * 3.33)
        self.decimal_places: int | None = (
            int(decimal_places) if decimal_places else None
        )
        self.base = base
        self.header: list[bytes] = []
        self.data: list[bytes] = []
        self.pen_table = PenTable(max_pens=256)
        self.current_pen: int = 0
        self.current_pen_width: float = 0.0
        self.setup(page)

        self._stroke_width_cache: dict[float, float] = dict()
        # StrokeWidthPolicy.absolute:
        # pen width in mm as resolved by the frontend
        self.min_lineweight = 0.05  # in mm, set by configure()
        self.lineweight_scaling = 1.0  # set by configure()
        self.lineweight_policy = LineweightPolicy.ABSOLUTE  # set by configure()
        # fixed lineweight for all strokes in ABSOLUTE mode:
        # set Configuration.min_lineweight to the desired lineweight in 1/300 inch!
        # set Configuration.lineweight_scaling to 0

        # LineweightPolicy.RELATIVE:
        # max_stroke_width is determined as a certain percentage of max(width, height)
        max_size = max(page.width_in_mm, page.height_in_mm)
        self.max_stroke_width: float = round(max_size * settings.max_stroke_width, 2)
        # min_stroke_width is determined as a certain percentage of max_stroke_width
        self.min_stroke_width: float = round(
            self.max_stroke_width * settings.min_stroke_width, 2
        )
        # LineweightPolicy.RELATIVE_FIXED:
        # all strokes have a fixed stroke-width as a certain percentage of max_stroke_width
        self.fixed_stroke_width: float = round(
            self.max_stroke_width * settings.fixed_stroke_width, 2
        )

    def setup(self, page: layout.Page) -> None:
        cmd = f"PS{page.width_in_mm*MM_TO_PLU:.0f},{page.height_in_mm*MM_TO_PLU:.0f};"
        self.header.append(cmd.encode())
        self.header.append(b"FT1;PA;")  # solid fill; plot absolute;

    def get_bytes(self) -> bytes:
        header: list[bytes] = list(self.header)
        header.append(self.pen_table.to_bytes())
        return compile_hpgl2(header, self.data)

    def switch_current_pen(self, pen_number: int, rgb: colors.RGB) -> int:
        if pen_number in self.pen_table:
            pen_color = self.pen_table[pen_number]
            if rgb != pen_color:
                self.data.append(make_pc(DEFAULT_PEN, rgb))
                pen_number = DEFAULT_PEN
        else:
            self.pen_table.add_pen(pen_number, rgb)
        return pen_number

    def set_pen(self, pen_number: int) -> None:
        if self.current_pen == pen_number:
            return
        self.data.append(f"SP{pen_number};".encode())
        self.current_pen = pen_number

    def set_pen_width(self, width: float) -> None:
        if self.current_pen_width == width:
            return
        self.data.append(f"PW{width:g};".encode())  # pen width in mm
        self.current_pen_width = width

    def enter_polygon_mode(self, start_point: Vec2) -> None:
        x = round(start_point.x, self.decimal_places)
        y = round(start_point.y, self.decimal_places)
        self.data.append(f"PA;PU{x},{y};PM;".encode())

    def close_current_polygon(self) -> None:
        self.data.append(b"PM1;")

    def fill_polygon(self) -> None:
        self.data.append(b"PM2;FP;")  # even/odd fill method

    def set_properties(self, properties: BackendProperties) -> None:
        pen_number = properties.pen
        pen_color, opacity = self.resolve_pen_color(properties.color)
        pen_width = self.resolve_pen_width(properties.lineweight)
        pen_number = self.switch_current_pen(pen_number, pen_color)
        self.set_pen(pen_number)
        self.set_pen_width(pen_width)

    def add_polyline_encoded(
        self, vertices: Iterable[Vec2], properties: BackendProperties
    ):
        self.set_properties(properties)
        self.data.append(polyline_encoder(vertices, self.factional_bits, self.base))

    def add_path(self, path: BkPath2d, properties: BackendProperties):
        if self.curves and path.has_curves:
            self.set_properties(properties)
            self.data.append(path_encoder(path, self.decimal_places))
        else:
            points = list(path.flattening(MAX_FLATTEN, segments=4))
            self.add_polyline_encoded(points, properties)

    @staticmethod
    def resolve_pen_color(color: Color) -> tuple[colors.RGB, float]:
        rgb = colors.RGB.from_hex(color)
        if rgb == WHITE:
            rgb = BLACK
        return rgb, alpha_to_opacity(color[7:9])

    def resolve_pen_width(self, width: float) -> float:
        try:
            return self._stroke_width_cache[width]
        except KeyError:
            pass
            stroke_width = self.fixed_stroke_width
        policy = self.lineweight_policy
        if policy == LineweightPolicy.ABSOLUTE:
            if self.lineweight_scaling:
                width = max(self.min_lineweight, width) * self.lineweight_scaling
            else:
                width = self.min_lineweight
            stroke_width = round(width, 2)  # in mm
        elif policy == LineweightPolicy.RELATIVE:
            stroke_width = round(
                map_lineweight_to_stroke_width(
                    width, self.min_stroke_width, self.max_stroke_width
                ),
                2,
            )
        self._stroke_width_cache[width] = stroke_width
        return stroke_width

    def set_background(self, color: Color) -> None:
        # background is always a white paper
        pass

    def draw_point(self, pos: Vec2, properties: BackendProperties) -> None:
        self.add_polyline_encoded([pos], properties)

    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties) -> None:
        self.add_polyline_encoded([start, end], properties)

    def draw_solid_lines(
        self, lines: Iterable[tuple[Vec2, Vec2]], properties: BackendProperties
    ) -> None:
        lines = list(lines)
        if len(lines) == 0:
            return
        for line in lines:
            self.add_polyline_encoded(line, properties)

    def draw_path(self, path: BkPath2d, properties: BackendProperties) -> None:
        for sub_path in path.sub_paths():
            if len(sub_path) == 0:
                continue
            self.add_path(sub_path, properties)

    def draw_filled_paths(
        self, paths: Iterable[BkPath2d], properties: BackendProperties
    ) -> None:
        paths = list(paths)
        if len(paths) == 0:
            return
        self.enter_polygon_mode(paths[0].start)
        for p in paths:
            for sub_path in p.sub_paths():
                if len(sub_path) == 0:
                    continue
                self.add_path(sub_path, properties)
                self.close_current_polygon()
        self.fill_polygon()

    def draw_filled_polygon(
        self, points: BkPoints2d, properties: BackendProperties
    ) -> None:
        points2d: list[Vec2] = points.vertices()
        if points2d:
            self.enter_polygon_mode(points2d[0])
            self.add_polyline_encoded(points2d, properties)
            self.fill_polygon()

    def draw_image(self, image_data: ImageData, properties: BackendProperties) -> None:
        pass  # TODO: not implemented

    def configure(self, config: Configuration) -> None:
        self.lineweight_policy = config.lineweight_policy
        if config.min_lineweight:
            # config.min_lineweight in 1/300 inch!
            min_lineweight_mm = config.min_lineweight * 25.4 / 300
            self.min_lineweight = max(0.05, min_lineweight_mm)
        self.lineweight_scaling = config.lineweight_scaling

    def clear(self) -> None:
        pass

    def finalize(self) -> None:
        pass

    def enter_entity(self, entity, properties) -> None:
        pass

    def exit_entity(self, entity) -> None:
        pass


def alpha_to_opacity(alpha: str) -> float:
    # stroke-opacity: 0.0 = transparent; 1.0 = opaque
    # alpha: "00" = transparent; "ff" = opaque
    if len(alpha):
        try:
            return int(alpha, 16) / 255
        except ValueError:
            pass
    return 1.0


def map_lineweight_to_stroke_width(
    lineweight: float,
    min_stroke_width: float,
    max_stroke_width: float,
    min_lineweight=0.05,  # defined by DXF
    max_lineweight=2.11,  # defined by DXF
) -> float:
    lineweight = max(min(lineweight, max_lineweight), min_lineweight) - min_lineweight
    factor = (max_stroke_width - min_stroke_width) / (max_lineweight - min_lineweight)
    return round(min_stroke_width + round(lineweight * factor), 2)


def flatten_path(path: BkPath2d) -> Sequence[Vec2]:
    points = list(path.flattening(distance=FLATTEN_MAX))
    return points


def compile_hpgl2(header: Sequence[bytes], commands: Sequence[bytes]) -> bytes:
    output = bytearray(PRELUDE)
    output.extend(b"".join(header))
    output.extend(b"".join(commands))
    output.extend(EPILOG)
    return bytes(output)


def pe_encode(value: float, frac_bits: int = 0, base: int = 64) -> bytes:
    if frac_bits:
        value *= 1 << frac_bits
    x = round(value)
    if x >= 0:
        x *= 2
    else:
        x = abs(x) * 2 + 1

    chars = bytearray()
    while x >= base:
        x, r = divmod(x, base)
        chars.append(63 + r)
    if base == 64:
        chars.append(191 + x)
    else:
        chars.append(95 + x)
    return bytes(chars)


def polyline_encoder(vertices: Iterable[Vec2], frac_bits: int, base: int) -> bytes:
    cmd = b"PE"
    if base == 32:
        cmd = b"PE7"
    if frac_bits:
        cmd += b">" + pe_encode(frac_bits, 0, base)
    data = [cmd + b"<="]
    vertices = list(vertices)
    # first point as absolute coordinates
    current = vertices[0]
    data.append(pe_encode(current.x, frac_bits, base))
    data.append(pe_encode(current.y, frac_bits, base))
    for vertex in vertices[1:]:
        # remaining points as relative coordinates
        delta = vertex - current
        data.append(pe_encode(delta.x, frac_bits, base))
        data.append(pe_encode(delta.y, frac_bits, base))
        current = vertex
    data.append(b";")
    return b"".join(data)


@no_type_check
def path_encoder(path: BkPath2d, decimal_places: int | None) -> bytes:
    # first point as absolute coordinates
    current = path.start
    x = round(current.x, decimal_places)
    y = round(current.y, decimal_places)
    data = [f"PU;PA{x:g},{y:g};PD;".encode()]
    prev_command = Command.MOVE_TO
    if len(path):
        commands: list[bytes] = []
        for cmd in path.commands():
            delta = cmd.end - current
            xe = round(delta.x, decimal_places)
            ye = round(delta.y, decimal_places)
            if cmd.type == Command.LINE_TO:
                coords = f"{xe:g},{ye:g};".encode()
                if prev_command == Command.LINE_TO:
                    # extend previous PR command
                    commands[-1] = commands[-1][:-1] + b"," + coords
                else:
                    commands.append(b"PR" + coords)
                prev_command = Command.LINE_TO
            else:
                if cmd.type == Command.CURVE3_TO:
                    control = cmd.ctrl - current
                    end = cmd.end - current
                    control_1 = 2.0 * control / 3.0
                    control_2 = end + 2.0 * (control - end) / 3.0
                elif cmd.type == Command.CURVE4_TO:
                    control_1 = cmd.ctrl1 - current
                    control_2 = cmd.ctrl2 - current
                else:
                    raise ValueError("internal error: MOVE_TO command is illegal here")
                x1 = round(control_1.x, decimal_places)
                y1 = round(control_1.y, decimal_places)
                x2 = round(control_2.x, decimal_places)
                y2 = round(control_2.y, decimal_places)
                coords = f"{x1:g},{y1:g},{x2:g},{y2:g},{xe:g},{ye:g};".encode()
                if prev_command == Command.CURVE4_TO:
                    # extend previous BR command
                    commands[-1] = commands[-1][:-1] + b"," + coords
                else:
                    commands.append(b"BR" + coords)
                prev_command = Command.CURVE4_TO
            current = cmd.end
        data.append(b"".join(commands))
    data.append(b"PU;")
    return b"".join(data)



# Copyright (c) 2019-2022 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf

from ezdxf.entities.dxfgfx import DXFGraphic, is_graphic_entity
from ezdxf.math import Matrix44
from ezdxf.lldxf.tags import Tags, DXFTag
from ezdxf.entities.dxfns import recover_graphic_attributes
from ezdxf.lldxf.const import DXFValueError
from ezdxf import colors


@pytest.fixture
def entity():
    return DXFGraphic.from_text(LINE)


def test_is_graphic_entity(entity):
    assert is_graphic_entity(entity) is True


def test_init_from_tags(entity):
    assert entity.dxf.layer == "Layer"


def test_true_color(entity):
    entity.dxf.true_color = 0x0F0F0F
    assert 0x0F0F0F == entity.dxf.true_color
    assert (
        0x0F,
        0x0F,
        0x0F,
    ) == entity.rgb  # shortcut for modern graphic entities
    entity.rgb = (255, 255, 255)  # shortcut for modern graphic entities
    assert 0xFFFFFF == entity.dxf.true_color


def test_color_name(entity):
    entity.dxf.color_name = "Rot"
    assert "Rot" == entity.dxf.color_name


def test_transparency(entity):
    entity.dxf.transparency = 0x020000FF  # 0xFF = opaque; 0x00 = 100% transparent
    assert 0x020000FF == entity.dxf.transparency
    # recommend usage: helper property ModernGraphicEntity.transparency
    assert 0.0 == entity.transparency  # 0. =  opaque; 1. = 100% transparent
    entity.transparency = 1.0
    assert 0x02000000 == entity.dxf.transparency


def test_default_attributes():
    entity = DXFGraphic.new()
    assert entity.dxf.layer == "0"
    assert entity.dxf.hasattr("layer") is True, "real attribute required"
    assert entity.dxf.color == 256
    assert entity.dxf.hasattr("color") is False, "just the default value"
    assert entity.dxf.linetype == "BYLAYER"
    assert entity.dxf.hasattr("linetype") is False, "just the default value"


def test_aci_color_index_fixer(entity):
    entity.dxf.color = -1
    assert entity.dxf.color == 256  # fixed as BYLAYER
    entity.dxf.color = 258
    assert entity.dxf.color == 256  # fixed as BYLAYER


def test_lineweight_fixer(entity):
    entity.dxf.lineweight = -5
    assert entity.dxf.lineweight == -1  # fixed as BYLAYER
    entity.dxf.lineweight = 17
    assert entity.dxf.lineweight == 18  # fixed as nearest valid lineweight
    entity.dxf.lineweight = 255
    assert entity.dxf.lineweight == 211  # fixed as nearest valid lineweight


def test_is_linetype_validator_active(entity):
    with pytest.raises(DXFValueError):
        entity.dxf.linetype = "*Invalid"


def test_is_layer_name_validator_active(entity):
    with pytest.raises(DXFValueError):
        entity.dxf.layer = "Invalid*"


def test_clone_graphical_entity(entity):
    entity.dxf.handle = "EFEF"
    entity.dxf.owner = "ABBA"
    entity.dxf.layer = "Layer1"
    entity.dxf.color = 13
    entity.set_reactors(["A", "F"])
    entity.set_xdata("MOZMAN", [(1000, "extended data")])

    clone = entity.copy()
    assert clone.dxf is not entity.dxf, "should be different objects"
    assert clone.dxf.handle is None, "should not have a handle"
    assert clone.dxf.owner is None
    assert clone.dxf.layer == "Layer1"
    assert clone.dxf.color == 13
    assert clone.reactors is not entity.reactors, "should be different objects"
    assert len(clone.get_reactors()) == 0
    assert clone.xdata is not entity.xdata, "should be different objects"
    assert clone.get_xdata("MOZMAN") == [(1000, "extended data")]

    clone.dxf.handle = "CDCD"
    clone.dxf.owner = "FEFE"
    clone.dxf.layer = "Layer2"
    clone.dxf.color = 77
    clone.set_reactors([])
    clone.set_xdata("MOZMAN", [(1000, "extended data1")])

    # don't change source entity
    assert entity.dxf.handle == "EFEF"
    assert entity.dxf.owner == "ABBA"
    assert entity.dxf.layer == "Layer1"
    assert entity.dxf.color == 13
    assert entity.get_reactors() == ["A", "F"]
    assert entity.get_xdata("MOZMAN") == [(1000, "extended data")]


def test_basic_transformation_interfaces():
    # test basic implementation = forward operation to transform interface
    class BasicGraphic(DXFGraphic):
        def transform(self, m: Matrix44) -> DXFGraphic:
            return self

    interface_mockup = BasicGraphic.new()
    assert interface_mockup.translate(1, 2, 3) is interface_mockup
    assert interface_mockup.scale(1, 2, 3) is interface_mockup
    assert interface_mockup.scale_uniform(1) is interface_mockup
    assert interface_mockup.rotate_axis((1, 2, 3), 1) is interface_mockup
    assert interface_mockup.rotate_x(1) is interface_mockup
    assert interface_mockup.rotate_y(1) is interface_mockup
    assert interface_mockup.rotate_z(1) is interface_mockup


def test_unlink_from_layout(entity):
    doc = ezdxf.new()
    msp = doc.modelspace()
    point = msp.add_point((0, 0))
    assert point.dxf.owner is not None
    point.unlink_from_layout()
    assert point.dxf.owner is None
    assert (
        point.dxf.handle in doc.entitydb
    ), "Do not delete unlinked entity from entitydb."

    # unlinking an already unlinked entity should pass silently
    point.unlink_from_layout()


def test_unlink_from_layout_without_doc(entity):
    entity.unlink_from_layout()
    assert entity.dxf.owner is None


def test_unlink_destroyed_entity_from_layout(entity):
    entity.destroy()
    with pytest.raises(TypeError):
        entity.unlink_from_layout()


def test_recover_acdb_entity_tags():
    entity = DXFGraphic()
    tags = Tags([DXFTag(62, 1), DXFTag(8, "Layer"), DXFTag(6, "Linetype")])

    recover_graphic_attributes(tags, entity.dxf)
    assert entity.dxf.color == 1
    assert entity.dxf.layer == "Layer"
    assert entity.dxf.linetype == "Linetype"


def test_recover_acdb_entity_tags_does_not_replace_existing_attribs():
    entity = DXFGraphic()
    entity.dxf.color = 7
    entity.dxf.layer = "HasLayer"
    entity.dxf.linetype = "HasLinetype"
    tags = Tags([DXFTag(62, 1), DXFTag(8, "Layer"), DXFTag(6, "Linetype")])

    recover_graphic_attributes(tags, entity.dxf)
    assert entity.dxf.color == 7
    assert entity.dxf.layer == "HasLayer"
    assert entity.dxf.linetype == "HasLinetype"


def test_recover_acdb_entity_tags_ignores_unknown_tags():
    entity = DXFGraphic()
    tags = Tags([DXFTag(62, 1), DXFTag(8, "Layer"), DXFTag(99, "Unknown")])

    unprocessed_tags = recover_graphic_attributes(tags, entity.dxf)
    assert len(unprocessed_tags) == 1
    assert unprocessed_tags[0] == (99, "Unknown")


class TestRrbAttribute:
    def test_set_rgb_from_tuple(self, entity: DXFGraphic):
        entity.rgb = (1, 2, 3)
        assert entity.dxf.true_color == 66051

    def test_set_rgb_from_rgb(self, entity: DXFGraphic):
        entity.rgb = colors.RGB(1, 2, 3)
        assert entity.rgb == (1, 2, 3)

    def test_setting_none_raises_type_error(self, entity: DXFGraphic):
        with pytest.raises(TypeError):
            entity.rgb = None

    def test_get_rgb(self, entity: DXFGraphic):
        entity.rgb = (1, 2, 3)
        assert entity.rgb == (1, 2, 3)

    def test_get_rgb_unset_true_color(self, entity: DXFGraphic):
        assert entity.rgb is None

    def test_delete_rgb(self, entity: DXFGraphic):
        entity.rgb = (1, 2, 3)
        del entity.rgb
        assert entity.dxf.hasattr("true_color") is False

    def test_deleting_unset_value_does_not_raise_exception(self, entity: DXFGraphic):
        del entity.rgb
        assert entity.dxf.hasattr("true_color") is False


if __name__ == "__main__":
    pytest.main([__file__])

LINE = """0
LINE
5
0
330
0
100
AcDbEntity
8
Layer
6
Linetype
62
77
370
25
100
AcDbLine
10
0.0
20
0.0
30
0.0
11
1.0
21
1.0
31
1.0
"""



# Copyright (c) 2018-2020 Manfred Moitzi
# License: MIT License
from typing import Iterable
import pytest
import math
from ezdxf.math.linalg import (
    Matrix,
    tridiagonal_vector_solver,
    tridiagonal_matrix_solver,
)
from ezdxf.math.legacy import (
    gauss_matrix_solver,
    gauss_vector_solver,
    gauss_jordan_solver,
    gauss_jordan_inverse,
    LUDecomposition,
)


@pytest.fixture
def X():
    return Matrix([[12, 7], [4, 5], [3, 8]])


def matrix_init(X):
    Y = Matrix([12, 7, 4, 5, 3, 8], shape=(3, 2))
    assert X.shape == X.nrows, X.ncols
    assert Y.shape == (3, 2)
    assert X == Y

    Y = Matrix(shape=(3, 3))
    assert Y == Matrix([[0, 0, 0], [0, 0, 0], [0, 0, 0]])

    Y = Matrix(X)
    assert Y == X
    Y[0, 0] = -1
    assert Y != X, "should not share the same list objects"

    Y = Matrix(X, shape=(2, 3))
    assert Y.rows() == [[12, 7, 4], [5, 3, 8]]


def test_matrix_getter(X):
    assert X[0, 0] == 12
    assert X[2, 1] == 8


def test_matrix_setter(X):
    X[0, 0] = -7
    X[2, 1] = 1.5
    assert X[0, 0] == -7
    assert X[2, 1] == 1.5


def test_row(X):
    a = list(X.row(0))
    assert list(X.row(0)) == [12, 7]
    assert list(X.row(1)) == [4, 5]
    assert list(X.row(2)) == [3, 8]
    assert list(X.rows()) == [[12, 7], [4, 5], [3, 8]]


def test_set_row(X):
    X.set_row(0, [1, 1])
    X.set_row(2, [2, 2])
    assert X.row(0) == [1, 1]
    assert X.row(2) == [2, 2]
    assert list(X.rows()) == [[1, 1], [4, 5], [2, 2]]

    X.set_row(-1, [7, 7])
    assert X.row(2) == [7, 7]


def test_set_row_error(X):
    with pytest.raises(IndexError):
        X.set_row(5, [1, 1])


def test_col(X):
    assert X.col(0) == [12, 4, 3]
    assert X.col(1) == [7, 5, 8]
    assert list(X.cols()) == [[12, 4, 3], [7, 5, 8]]


def test_set_col(X):
    X.set_col(0, [1, 1, 1])
    X.set_col(1, [2, 2, 2])
    assert X.col(0) == [1, 1, 1]
    assert X.col(1) == [2, 2, 2]
    assert list(X.cols()) == [[1, 1, 1], [2, 2, 2]]

    X.set_col(-1, [3, 3, 3])
    assert X.col(1) == [3, 3, 3]


def test_set_col_error(X):
    with pytest.raises(IndexError):
        X.set_col(2, [1, 1, 1])


def test_freeze_matrix(X):
    m = X.freeze()
    assert m == X
    assert m[0, 0] == 12
    assert m[2, 1] == 8
    with pytest.raises(ValueError):
        m[0, 0] = 1.0


def test_mul():
    X = Matrix(
        [
            [12, 7, 3],
            [4, 5, 6],
            [7, 8, 9],
        ]
    )

    Y = Matrix(
        [
            [5, 8, 1, 2],
            [6, 7, 3, 0],
            [4, 5, 9, 1],
        ]
    )

    R = Matrix(
        [
            [114, 160, 60, 27],
            [74, 97, 73, 14],
            [119, 157, 112, 23],
        ]
    )

    assert X * Y == R


def test_imul():
    X = Matrix(
        [
            [12, 7, 3],
            [4, 5, 6],
            [7, 8, 9],
        ]
    )
    Y = X
    Y *= 10

    assert Y == Matrix(
        [
            [120, 70, 30],
            [40, 50, 60],
            [70, 80, 90],
        ]
    )
    assert X[0, 0] != Y[0, 0]


def test_transpose(X):
    R = Matrix(
        [
            (12, 4, 3),
            (7, 5, 8),
        ]
    )
    T = X.transpose()
    assert T == R
    # is T mutable?
    T[0, 0] = 99
    assert T[0, 0] == 99


def test_add(X):
    R = X + X
    assert R == Matrix([[24, 14], [8, 10], [6, 16]])

    R = R + 10
    assert R == Matrix([[34, 24], [18, 20], [16, 26]])


def test_iadd(X):
    Y = X
    Y += Y
    assert Y == Matrix([[24, 14], [8, 10], [6, 16]])
    assert Y[0, 0] != X[0, 0]
    Z = Y
    Z += 10
    assert Z == Matrix([[34, 24], [18, 20], [16, 26]])
    assert Y[0, 0] != Z[0, 0]


def test_sub(X):
    R = X - X
    assert R == Matrix([[0, 0], [0, 0], [0, 0]])

    R = X - 10
    assert R == Matrix([[2, -3], [-6, -5], [-7, -2]])


def test_build_matrix_by_rows():
    m = Matrix()
    m.append_row([1, 2, 3])
    assert m.nrows == 1
    assert m.ncols == 3


def test_build_matrix_by_cols():
    m = Matrix()
    m.append_col([1, 2, 3])
    assert m.nrows == 3
    assert m.ncols == 1


DIAG = [
    [1, 2, 5, 7],
    [5, 2, 3, 6],
    [7, 4, 3, 4],
    [8, 6, 3, 4],
]


def test_diag():
    m = Matrix(DIAG)
    assert m.diag(0) == [1, 2, 3, 4]
    assert m.diag(1) == [2, 3, 4]
    assert m.diag(2) == [5, 6]
    assert m.diag(3) == [7]
    assert m.diag(4) == []
    assert m.diag(-1) == [5, 4, 3]
    assert m.diag(-2) == [7, 6]
    assert m.diag(-3) == [8]
    assert m.diag(-4) == []


def test_set_diag_float():
    m = Matrix(shape=(4, 4))
    m.set_diag(0, 2)
    for i in range(4):
        assert m[i, i] == 2.0


def test_set_diag_above():
    m = Matrix(shape=(4, 4))
    m.set_diag(1, 2)
    for i in range(3):
        assert m[i, i + 1] == 2.0


def test_set_diag_below():
    m = Matrix(shape=(4, 4))
    m.set_diag(-1, 2)
    for i in range(3):
        assert m[i + 1, i] == 2.0


def test_set_diag_iterable():
    m = Matrix(shape=(4, 4))
    m.set_diag(0, range(5))
    for i in range(4):
        assert m[i, i] == i


def test_identity():
    m = Matrix.identity((3, 4))
    for i in range(3):
        assert m[i, i] == 1.0


A = [
    [2, 3, 2, 5, 6],
    [5, 1, 4, 5, 3],
    [1, 12, 3, 1, 12],
    [7, 3, 2, 2, 6],
    [9, 4, 2, 13, 6],
]
B1 = [6, 9, 5, 4, 8]
B2 = [5, 10, 6, 3, 2]
B3 = [1, 7, 3, 9, 12]

SOLUTION_B1 = [
    -0.14854771784232382,
    -0.3128630705394192,
    1.7966804979253113,
    0.41908713692946065,
    0.2578146611341633,
]


def test_gauss_vector_solver():
    result = gauss_vector_solver(A, B1)
    assert result == SOLUTION_B1


def is_close_vectors(v0, v1) -> bool:
    return all(math.isclose(c0, c1) for c0, c1 in zip(v0, v1))


def test_gauss_matrix_solver():
    result = gauss_matrix_solver(A, zip(B1, B2, B3))
    assert is_close_vectors(result.col(0), gauss_vector_solver(A, B1))
    assert is_close_vectors(result.col(1), gauss_vector_solver(A, B2))
    assert is_close_vectors(result.col(2), gauss_vector_solver(A, B3))


def are_close_vectors(v1: Iterable[float], v2: Iterable[float], abs_tol: float = 1e-12):
    for i, j in zip(v1, v2):
        assert math.isclose(i, j, abs_tol=abs_tol)


def test_gauss_jordan_vector_solver():
    B = Matrix(items=B1, shape=(5, 1))
    result_A, result_B = gauss_jordan_solver(A, B)
    are_close_vectors(result_B.col(0), SOLUTION_B1)


def test_gauss_jordan_matrix_solver():
    result_A, result_B = gauss_jordan_solver(A, zip(B1, B2, B3))
    are_close_vectors(result_B.col(0), gauss_vector_solver(A, B1))
    are_close_vectors(result_B.col(1), gauss_vector_solver(A, B2))
    are_close_vectors(result_B.col(2), gauss_vector_solver(A, B3))


EXPECTED_INVERSE = [
    [
        -0.16390041493775933617,
        -0.002489626556016597522,
        -0.007468879668049792526,
        0.13651452282157676342,
        0.043568464730290456478,
    ],
    [
        -0.33402489626556016593,
        0.05062240663900414948,
        0.15186721991701244817,
        -0.10912863070539419082,
        0.11410788381742738587,
    ],
    [
        -0.05394190871369294633,
        0.34854771784232365142,
        0.04564315352697095431,
        -0.11203319502074688787,
        -0.099585062240663900493,
    ],
    [
        0.06016597510373443986,
        -0.00414937759336099577,
        -0.012448132780082987519,
        -0.10580912863070539416,
        0.072614107883817427371,
    ],
    [
        0.35615491009681881048,
        -0.13720608575380359624,
        -0.078284923928077455093,
        0.13457814661134163205,
        -0.098893499308437067754,
    ],
]


def test_gauss_jordan_inverse():
    result = gauss_jordan_inverse(A)
    assert result.nrows == len(A)
    assert result.ncols == len(A[0])
    m = Matrix(matrix=EXPECTED_INVERSE)
    assert result.isclose(m)


def test_LU_decomposition_solve_vector():
    R = LUDecomposition(A).solve_vector(B1)
    are_close_vectors(R, SOLUTION_B1)


def test_LU_decomposition_solve_matrix():
    lu = LUDecomposition(A)
    result = lu.solve_matrix(zip(B1, B2, B3))
    are_close_vectors(result.col(0), gauss_vector_solver(A, B1))
    are_close_vectors(result.col(1), gauss_vector_solver(A, B2))
    are_close_vectors(result.col(2), gauss_vector_solver(A, B3))


def test_LU_decomposition_inverse():
    m = Matrix(matrix=EXPECTED_INVERSE)
    assert m.isclose(LUDecomposition(A).inverse())


def test_determinant():
    from ezdxf.math import Matrix44

    A = [
        [2, 3, 2, 5],
        [5, 1, 4, 5],
        [1, 12, 3, 1],
        [7, 3, 2, 2],
    ]
    det = LUDecomposition(A).determinant()
    chk = Matrix44(*A)
    assert math.isclose(chk.determinant(), det)


TRI_DIAGONAL = [
    [2, 3, 0, 0, 0],
    [5, 1, 4, 0, 0],
    [0, 9, 3, 1, 0],
    [0, 0, 2, 2, 6],
    [0, 0, 0, 4, 6],
]


def tri_solution():
    return gauss_matrix_solver(TRI_DIAGONAL, zip(B1, B2, B3))


@pytest.fixture
def tridiag():
    m = Matrix(TRI_DIAGONAL)
    a = [0]  # a0 is not used but must be present
    a.extend(m.diag(-1))
    b = m.diag(0)
    c = m.diag(+1)
    return a, b, c


def test_tridiagonal_vector_solver(tridiag):
    result = tridiagonal_vector_solver(tridiag, B1)
    are_close_vectors(result, tri_solution().col(0))


def test_tridiagonal_matrix_solver(tridiag):
    result = tridiagonal_matrix_solver(tridiag, zip(B1, B2, B3))
    assert result.isclose(tri_solution())



#  Copyright (c) 2020-2024, Manfred Moitzi
#  License: MIT License
from pathlib import Path
import pytest
import ezdxf
from ezdxf import recover

DATADIR = Path(__file__).parent / "data"


def test_read_sublass_markers_in_r12_entities():
    """DXF R12 file with subclass markers (100, ...) in DXF entities."""
    doc = ezdxf.readfile(DATADIR / "Leica_Disto_S910.dxf")
    msp = doc.modelspace()
    points = list(msp.query("POINT"))
    assert len(points) == 11
    assert len(points[0].dxf.location) == 3


def test_read_sublass_markers_in_r12_table_entries():
    """DXF R12 file with subclass markers (100, ...) in table entries."""
    doc, _ = recover.readfile(DATADIR / "issue1106.dxf")
    msp = doc.modelspace()
    assert len(msp) == 10


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2021-2025, Manfred Moitzi
# License: MIT License

# This is the abstract link between the text layout engine implemented in
# ezdxf.tools.text_layout and a concrete MTEXT renderer implementation like
# MTextExplode or ComplexMTextRenderer.
from __future__ import annotations
from typing import Sequence, Optional
import abc
from ezdxf.lldxf import const
from ezdxf import colors
from ezdxf.entities.mtext import MText, MTextColumns
from ezdxf.entities.textstyle import get_textstyle
from ezdxf.enums import (
    MTextParagraphAlignment,
)
from ezdxf.fonts import fonts
from ezdxf.tools import text_layout as tl
from ezdxf.tools.text import (
    MTextParser,
    MTextContext,
    TokenType,
    ParagraphProperties,
    estimate_mtext_extents,
    valid_text_height,
)

__all__ = ["AbstractMTextRenderer"]

ALIGN = {
    MTextParagraphAlignment.LEFT: tl.ParagraphAlignment.LEFT,
    MTextParagraphAlignment.RIGHT: tl.ParagraphAlignment.RIGHT,
    MTextParagraphAlignment.CENTER: tl.ParagraphAlignment.CENTER,
    MTextParagraphAlignment.JUSTIFIED: tl.ParagraphAlignment.JUSTIFIED,
    MTextParagraphAlignment.DISTRIBUTED: tl.ParagraphAlignment.JUSTIFIED,
    MTextParagraphAlignment.DEFAULT: tl.ParagraphAlignment.LEFT,
}

ATTACHMENT_POINT_TO_ALIGN = {
    const.MTEXT_TOP_LEFT: tl.ParagraphAlignment.LEFT,
    const.MTEXT_MIDDLE_LEFT: tl.ParagraphAlignment.LEFT,
    const.MTEXT_BOTTOM_LEFT: tl.ParagraphAlignment.LEFT,
    const.MTEXT_TOP_CENTER: tl.ParagraphAlignment.CENTER,
    const.MTEXT_MIDDLE_CENTER: tl.ParagraphAlignment.CENTER,
    const.MTEXT_BOTTOM_CENTER: tl.ParagraphAlignment.CENTER,
    const.MTEXT_TOP_RIGHT: tl.ParagraphAlignment.RIGHT,
    const.MTEXT_MIDDLE_RIGHT: tl.ParagraphAlignment.RIGHT,
    const.MTEXT_BOTTOM_RIGHT: tl.ParagraphAlignment.RIGHT,
}

STACKING = {
    "^": tl.Stacking.OVER,
    "/": tl.Stacking.LINE,
    "#": tl.Stacking.SLANTED,
}


def make_default_tab_stops(cap_height: float, width: float) -> list[tl.TabStop]:
    tab_stops = []
    cap_height = valid_text_height(cap_height)
    step = 4.0 * cap_height
    pos = step
    while pos < width:
        tab_stops.append(tl.TabStop(pos, tl.TabStopType.LEFT))
        pos += step
    return tab_stops


def append_default_tab_stops(
    tab_stops: list[tl.TabStop], default_stops: Sequence[tl.TabStop]
) -> None:
    last_pos = 0.0
    if tab_stops:
        last_pos = tab_stops[-1].pos
    tab_stops.extend(stop for stop in default_stops if stop.pos > last_pos)


def make_tab_stops(
    cap_height: float,
    width: float,
    tab_stops: Sequence,
    default_stops: Sequence[tl.TabStop],
) -> list[tl.TabStop]:
    cap_height = valid_text_height(cap_height)
    _tab_stops = []
    for stop in tab_stops:
        if isinstance(stop, str):
            value = float(stop[1:])
            if stop[0] == "c":
                kind = tl.TabStopType.CENTER
            else:
                kind = tl.TabStopType.RIGHT
        else:
            kind = tl.TabStopType.LEFT
            value = float(stop)
        pos = value * cap_height
        if pos < width:
            _tab_stops.append(tl.TabStop(pos, kind))

    append_default_tab_stops(_tab_stops, default_stops)
    return _tab_stops


def get_stroke(ctx: MTextContext) -> int:
    stroke = 0
    if ctx.underline:
        stroke += tl.Stroke.UNDERLINE
    if ctx.strike_through:
        stroke += tl.Stroke.STRIKE_THROUGH
    if ctx.overline:
        stroke += tl.Stroke.OVERLINE
    if ctx.continue_stroke:
        stroke += tl.Stroke.CONTINUE
    return stroke


def new_paragraph(
    cells: list,
    ctx: MTextContext,
    cap_height: float,
    line_spacing: float = 1,
    width: float = 0,
    default_stops: Optional[Sequence[tl.TabStop]] = None,
):
    cap_height = valid_text_height(cap_height)
    if cells:
        p = ctx.paragraph
        align = ALIGN.get(p.align, tl.ParagraphAlignment.LEFT)
        left = p.left * cap_height
        right = p.right * cap_height
        first = left + p.indent * cap_height  # relative to left
        _default_stops: Sequence[tl.TabStop] = default_stops or []
        tab_stops = _default_stops
        if p.tab_stops:
            tab_stops = make_tab_stops(cap_height, width, p.tab_stops, _default_stops)
        paragraph = tl.Paragraph(
            align=align,
            indent=(first, left, right),
            line_spacing=line_spacing,
            tab_stops=tab_stops,
        )
        paragraph.append_content(cells)
    else:
        paragraph = tl.EmptyParagraph(  # type: ignore
            cap_height=ctx.cap_height, line_spacing=line_spacing
        )
    return paragraph


def super_glue():
    return tl.NonBreakingSpace(width=0, min_width=0, max_width=0)


def defined_width(mtext: MText) -> float:
    width = mtext.dxf.get("width", 0.0)
    if width < 1e-6:
        width, height = estimate_mtext_extents(mtext)
    return width


def column_heights(columns: MTextColumns) -> list[Optional[float]]:
    heights: list[Optional[float]]
    if columns.heights:  # dynamic manual
        heights = list(columns.heights)
        # last height has to be auto height = None
        heights[-1] = None
        return heights
    # static, dynamic auto
    defined_height = abs(columns.defined_height)
    if defined_height < 1e-6:
        return [None]
    return [defined_height] * columns.count


class AbstractMTextRenderer(abc.ABC):
    def __init__(self) -> None:
        self._font_cache: dict[tuple[str, float, float], fonts.AbstractFont] = {}

    @abc.abstractmethod
    def word(self, test: str, ctx: MTextContext) -> tl.ContentCell: ...

    @abc.abstractmethod
    def fraction(
        self, data: tuple[str, str, str], ctx: MTextContext
    ) -> tl.ContentCell: ...

    @abc.abstractmethod
    def get_font_face(self, mtext: MText) -> fonts.FontFace: ...

    @abc.abstractmethod
    def make_bg_renderer(self, mtext: MText) -> tl.ContentRenderer: ...

    def make_mtext_context(self, mtext: MText) -> MTextContext:
        ctx = MTextContext()
        ctx.paragraph = ParagraphProperties(
            align=ATTACHMENT_POINT_TO_ALIGN.get(  # type: ignore
                mtext.dxf.attachment_point, tl.ParagraphAlignment.LEFT
            )
        )
        ctx.font_face = self.get_font_face(mtext)
        ctx.cap_height = valid_text_height(mtext.dxf.char_height)
        ctx.aci = mtext.dxf.color
        textstyle = get_textstyle(mtext)
        ctx.width_factor = textstyle.dxf.width
        ctx.oblique = textstyle.dxf.oblique
        rgb = mtext.rgb
        if rgb is not None:
            ctx.rgb = colors.RGB(*rgb)
        return ctx

    def get_font(self, ctx: MTextContext) -> fonts.AbstractFont:
        ttf = fonts.find_font_file_name(ctx.font_face)  # 1st call is very slow
        key = (ttf, ctx.cap_height, ctx.width_factor)
        font = self._font_cache.get(key)
        if font is None:
            font = fonts.make_font(ttf, ctx.cap_height, ctx.width_factor)
            self._font_cache[key] = font
        return font

    def get_stroke(self, ctx: MTextContext) -> int:
        return get_stroke(ctx)

    def get_stacking(self, type_: str) -> tl.Stacking:
        return STACKING.get(type_, tl.Stacking.LINE)

    def space_width(self, ctx: MTextContext) -> float:
        return self.get_font(ctx).space_width()

    def space(self, ctx: MTextContext):
        return tl.Space(width=self.space_width(ctx))

    def tabulator(self, ctx: MTextContext):
        return tl.Tabulator(width=self.space_width(ctx))

    def non_breaking_space(self, ctx: MTextContext):
        return tl.NonBreakingSpace(width=self.space_width(ctx))

    def layout_engine(self, mtext: MText) -> tl.Layout:
        initial_cap_height = valid_text_height(mtext.dxf.char_height)
        line_spacing = mtext.dxf.line_spacing_factor

        def append_paragraph():
            paragraph = new_paragraph(
                cells,
                ctx,
                initial_cap_height,
                line_spacing,
                width,
                default_stops,
            )
            layout.append_paragraphs([paragraph])
            cells.clear()

        bg_renderer = self.make_bg_renderer(mtext)
        width = defined_width(mtext)
        default_stops = make_default_tab_stops(initial_cap_height, width)
        layout = tl.Layout(width=width)
        if mtext.has_columns:
            columns = mtext.columns
            assert columns is not None
            for height in column_heights(columns):
                layout.append_column(
                    width=columns.width,
                    height=height,
                    gutter=columns.gutter_width,
                    renderer=bg_renderer,
                )
        else:
            # column with auto height and default width
            layout.append_column(renderer=bg_renderer)

        content = mtext.all_columns_raw_content()
        ctx = self.make_mtext_context(mtext)
        cells: list[tl.Cell] = []
        for token in MTextParser(content, ctx):
            ctx = token.ctx
            if token.type == TokenType.NEW_PARAGRAPH:
                append_paragraph()
            elif token.type == TokenType.NEW_COLUMN:
                append_paragraph()
                layout.next_column()
            elif token.type == TokenType.SPACE:
                cells.append(self.space(ctx))
            elif token.type == TokenType.NBSP:
                cells.append(self.non_breaking_space(ctx))
            elif token.type == TokenType.TABULATOR:
                cells.append(self.tabulator(ctx))
            elif token.type == TokenType.WORD:
                if cells and isinstance(cells[-1], (tl.Text, tl.Fraction)):
                    # Create an unbreakable connection between those two parts.
                    cells.append(super_glue())
                cells.append(self.word(token.data, ctx))
            elif token.type == TokenType.STACK:
                if cells and isinstance(cells[-1], (tl.Text, tl.Fraction)):
                    # Create an unbreakable connection between those two parts.
                    cells.append(super_glue())
                cells.append(self.fraction(token.data, ctx))

        if cells:
            append_paragraph()

        return layout



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math import (
    intersect_polylines_2d,
    intersect_polylines_3d,
    Vec2,
    Vec3,
    ConstructionEllipse,
    BSpline,
)
from ezdxf.render import forms


class TestIntersectPolylines2d:
    def test_intersecting_single_segments(self):
        pline1 = Vec2.list([(0, 1), (2, 1)])
        pline2 = Vec2.list([(1, 0), (1, 2)])
        res = intersect_polylines_2d(pline1, pline2)
        assert len(res) == 1
        assert res[0].isclose(Vec2(1, 1))

    def test_none_intersecting_single_segments(self):
        pline1 = Vec2.list([(0, 0), (2, 0)])
        pline2 = Vec2.list([(0, 1), (2, 1)])
        res = intersect_polylines_2d(pline1, pline2)
        assert len(res) == 0

    def test_intersecting_cross(self):
        pline1 = Vec2.list([(0, 1), (1, 1), (2, 1)])
        pline2 = Vec2.list([(1, 0), (1, 1), (1, 2)])
        res = intersect_polylines_2d(pline1, pline2)
        assert len(res) == 1
        assert res[0].isclose(Vec2(1, 1))

    def test_intersecting_x_cross(self):
        pline1 = Vec2.list([(0, 0), (1, 1), (2, 2)])
        pline2 = Vec2.list([(2, 0), (1, 1), (0, 2)])
        res = intersect_polylines_2d(pline1, pline2)
        assert len(res) == 1
        assert res[0].isclose(Vec2(1, 1))

    def test_intersecting_zig_zag_lines(self):
        pline1 = Vec2.list([(0, 0), (2, 2), (4, 0), (6, 2), (8, 0)])
        pline2 = Vec2.list([(0, 2), (2, 0), (4, 2), (6, 0), (8, 2)])
        res = intersect_polylines_2d(pline1, pline2)
        assert len(res) == 4
        res.sort()  # do not rely on any order
        assert res[0].isclose(Vec2(1, 1))
        assert res[1].isclose(Vec2(3, 1))
        assert res[2].isclose(Vec2(5, 1))
        assert res[3].isclose(Vec2(7, 1))

    def test_zig_zag_lines_with_common_vertices(self):
        pline1 = Vec2.list([(0, 0), (2, 2), (4, 0), (6, 2), (8, 0)])
        pline2 = Vec2.list([(0, 4), (2, 2), (4, 4), (6, 2), (8, 4)])
        res = intersect_polylines_2d(pline1, pline2)
        assert len(res) == 2
        res.sort()  # do not rely on any order
        assert res[0].isclose(Vec2(2, 2))
        assert res[1].isclose(Vec2(6, 2))

    def test_complex_ellipse_with_spline_intersection(self):
        ellipse = ConstructionEllipse(center=(0, 0), major_axis=(3, 0), ratio=0.5)
        bspline = BSpline([(-4, -4), (-2, -1), (2, 1), (4, 4)])
        p1 = ellipse.flattening(distance=0.01)
        p2 = bspline.flattening(distance=0.01)
        res = intersect_polylines_2d(Vec2.list(p1), Vec2.list(p2))
        assert len(res) == 2

    def test_intersecting_squares(self):
        square1 = forms.close_polygon(forms.square(2.0))
        square2 = forms.translate(square1, (1, 1))
        res = intersect_polylines_2d(Vec2.list(square1), Vec2.list(square2))
        assert len(res) == 2
        res.sort()
        assert res[0].isclose(Vec2(1, 2))
        assert res[1].isclose(Vec2(2, 1))

    def test_squares_with_common_corner_vertex(self):
        square1 = forms.close_polygon(forms.square(2.0))
        square2 = forms.translate(square1, (2, 2))
        res = intersect_polylines_2d(Vec2.list(square1), Vec2.list(square2))
        assert len(res) == 1
        assert res[0].isclose(Vec2(2, 2))

    def test_coincident_common_segment(self):
        """ The common segment does not create intersection points.
        Same as intersection of coincident lines!
        """
        pline1 = Vec2.list([(1, 1), (2, 1)])
        pline2 = Vec2.list([(1, 1), (2, 1)])
        res = intersect_polylines_2d(pline1, pline2)
        assert len(res) == 0

    def test_coincident_common_last_segment(self):
        """ The common segment does not create intersection points, but the
        preceding segment does.
        """
        pline1 = Vec2.list([(0, 0), (1, 1), (2, 1)])
        pline2 = Vec2.list([(0, 2), (1, 1), (2, 1)])
        res = intersect_polylines_2d(pline1, pline2)
        assert len(res) == 1
        assert res[0].isclose(Vec2(1, 1))

    def test_coincident_common_intermediate_segment(self):
        """ The common segment does not create intersection points, but the
        preceding and the following segment does.
        """
        pline1 = Vec2.list([(0, 0), (1, 1), (2, 1), (3, 0)])
        pline2 = Vec2.list([(0, 2), (1, 1), (2, 1), (3, 2)])
        res = intersect_polylines_2d(pline1, pline2)
        assert len(res) == 2
        res.sort()
        assert res[0].isclose(Vec2(1, 1))
        assert res[1].isclose(Vec2(2, 1))


class TestIntersectPolylines3d:
    """
    The 3D line intersection function is tested in test suite 614.

    The basic polyline intersection algorithm is tested in class TestIntersectPolylines2d.

    This class tests only if the function intersect_polylines_3d() is implemented.

    """
    def test_intersecting_single_segments(self):
        pline1 = Vec3.list([(0, 1), (2, 1)])
        pline2 = Vec3.list([(1, 0), (1, 2)])
        res = intersect_polylines_3d(pline1, pline2)
        assert len(res) == 1
        assert res[0].isclose(Vec3(1, 1))

    def test_intersecting_single_vertical_segment(self):
        pline1 = Vec3.list([(0, 0, 0), (2, 0, 0)])
        pline2 = Vec3.list([(1, 0, 0), (1, 0, 1)])
        res = intersect_polylines_3d(pline1, pline2)
        assert len(res) == 1
        assert res[0].isclose(Vec3(1, 0))

    def test_none_intersecting_single_segments(self):
        pline1 = Vec3.list([(0, 0), (2, 0)])
        pline2 = Vec3.list([(0, 1), (2, 1)])
        res = intersect_polylines_3d(pline1, pline2)
        assert len(res) == 0


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterable, Sequence, Optional, Iterator, Callable
from typing_extensions import Protocol
import math
import enum

from ezdxf.math import (
    Vec2,
    UVec,
    intersection_line_line_2d,
    is_point_in_polygon_2d,
    has_clockwise_orientation,
    point_to_line_relation,
    TOLERANCE,
    BoundingBox2d,
)
from ezdxf.tools import take2, pairwise


__all__ = [
    "greiner_hormann_union",
    "greiner_hormann_difference",
    "greiner_hormann_intersection",
    "Clipping",
    "ConvexClippingPolygon2d",
    "ConcaveClippingPolygon2d",
    "ClippingRect2d",
    "InvertedClippingPolygon2d",
    "CohenSutherlandLineClipping2d",
]


class Clipping(Protocol):
    def clip_polygon(self, polygon: Sequence[Vec2]) -> Sequence[Sequence[Vec2]]:
        """Returns the parts of the clipped polygon."""

    def clip_polyline(self, polyline: Sequence[Vec2]) -> Sequence[Sequence[Vec2]]:
        """Returns the parts of the clipped polyline."""

    def clip_line(self, start: Vec2, end: Vec2) -> Sequence[tuple[Vec2, Vec2]]:
        """Returns the parts of the clipped line."""

    def is_inside(self, point: Vec2) -> bool:
        """Returns ``True`` if `point` is inside the clipping path."""


def _clip_polyline(
    polyline: Sequence[Vec2],
    line_clipper: Callable[[Vec2, Vec2], Sequence[tuple[Vec2, Vec2]]],
    abs_tol: float,
) -> Sequence[Sequence[Vec2]]:
    """Returns the parts of the clipped polyline."""
    if len(polyline) < 2:
        return []
    result: list[Vec2] = []
    parts: list[list[Vec2]] = []
    next_start = polyline[0]
    for end in polyline[1:]:
        start = next_start
        next_start = end
        for clipped_line in line_clipper(start, end):
            if len(clipped_line) != 2:
                continue
            if result:
                clip_start, clip_end = clipped_line
                if result[-1].isclose(clip_start, abs_tol=abs_tol):
                    result.append(clip_end)
                    continue
                parts.append(result)
            result = list(clipped_line)
    if result:
        parts.append(result)
    return parts


class ConvexClippingPolygon2d:
    """The clipping path is an arbitrary convex 2D polygon."""

    def __init__(self, vertices: Iterable[Vec2], ccw_check=True, abs_tol=TOLERANCE):
        self.abs_tol = abs_tol
        clip = list(vertices)
        if len(clip) > 1:
            if clip[0].isclose(clip[-1], abs_tol=self.abs_tol):
                clip.pop()
        if len(clip) < 3:
            raise ValueError("more than 3 vertices as clipping polygon required")
        if ccw_check and has_clockwise_orientation(clip):
            clip.reverse()
        self._clipping_polygon: list[Vec2] = clip

    def clip_polyline(self, polyline: Sequence[Vec2]) -> Sequence[Sequence[Vec2]]:
        """Returns the parts of the clipped polyline."""
        return _clip_polyline(polyline, self.clip_line, abs_tol=self.abs_tol)

    def clip_line(self, start: Vec2, end: Vec2) -> Sequence[tuple[Vec2, Vec2]]:
        """Returns the parts of the clipped line."""

        def is_inside(point: Vec2) -> bool:
            # is point left of line:
            return (clip_end.x - clip_start.x) * (point.y - clip_start.y) - (
                clip_end.y - clip_start.y
            ) * (point.x - clip_start.x) >= 0.0

        def edge_intersection(default: Vec2) -> Vec2:
            ip = intersection_line_line_2d(
                (edge_start, edge_end), (clip_start, clip_end), abs_tol=self.abs_tol
            )
            if ip is None:
                return default
            return ip

        # The clipping polygon is always treated as a closed polyline!
        clip_start = self._clipping_polygon[-1]
        edge_start = start
        edge_end = end
        for clip_end in self._clipping_polygon:
            if is_inside(edge_start):
                if not is_inside(edge_end):
                    edge_end = edge_intersection(edge_end)
            elif is_inside(edge_end):
                if not is_inside(edge_start):
                    edge_start = edge_intersection(edge_start)
            else:
                return tuple()
            clip_start = clip_end
        return ((edge_start, edge_end),)

    def clip_polygon(self, polygon: Sequence[Vec2]) -> Sequence[Sequence[Vec2]]:
        """Returns the parts of the clipped polygon. A polygon is a closed polyline."""

        def is_inside(point: Vec2) -> bool:
            # is point left of line:
            return (clip_end.x - clip_start.x) * (point.y - clip_start.y) - (
                clip_end.y - clip_start.y
            ) * (point.x - clip_start.x) > 0.0

        def edge_intersection() -> None:
            ip = intersection_line_line_2d(
                (edge_start, edge_end), (clip_start, clip_end), abs_tol=self.abs_tol
            )
            if ip is not None:
                clipped.append(ip)

        # The clipping polygon is always treated as a closed polyline!
        clip_start = self._clipping_polygon[-1]
        clipped = list(polygon)
        for clip_end in self._clipping_polygon:
            # next clipping edge to test: clip_start -> clip_end
            if not clipped:  # no subject vertices left to test
                break

            vertices = clipped.copy()
            if len(vertices) > 1 and vertices[0].isclose(
                vertices[-1], abs_tol=self.abs_tol
            ):
                vertices.pop()

            clipped.clear()
            edge_start = vertices[-1]
            for edge_end in vertices:
                # next polygon edge to test: edge_start -> edge_end
                if is_inside(edge_end):
                    if not is_inside(edge_start):
                        edge_intersection()
                    clipped.append(edge_end)
                elif is_inside(edge_start):
                    edge_intersection()
                edge_start = edge_end
            clip_start = clip_end
        return (clipped,)

    def is_inside(self, point: Vec2) -> bool:
        """Returns ``True`` if `point` is inside the clipping polygon."""
        return is_point_in_polygon_2d(point, self._clipping_polygon) >= 0


class ClippingRect2d:
    """The clipping path is an axis-aligned rectangle, where all sides are parallel to
    the x- and y-axis.
    """

    def __init__(self, bottom_left: Vec2, top_right: Vec2, abs_tol=TOLERANCE):
        self.abs_tol = abs_tol
        self._bbox = BoundingBox2d((bottom_left, top_right))
        bottom_left = self._bbox.extmin
        top_right = self._bbox.extmax
        self._clipping_polygon = ConvexClippingPolygon2d(
            [
                bottom_left,
                Vec2(top_right.x, bottom_left.y),
                top_right,
                Vec2(bottom_left.x, top_right.y),
            ],
            ccw_check=False,
            abs_tol=self.abs_tol,
        )
        self._line_clipper = CohenSutherlandLineClipping2d(
            self._bbox.extmin, self._bbox.extmax
        )

    def clip_polygon(self, polygon: Sequence[Vec2]) -> Sequence[Sequence[Vec2]]:
        """Returns the parts of the clipped polygon. A polygon is a closed polyline."""
        return self._clipping_polygon.clip_polygon(polygon)

    def clip_polyline(self, polyline: Sequence[Vec2]) -> Sequence[Sequence[Vec2]]:
        """Returns the parts of the clipped polyline."""
        return _clip_polyline(polyline, self.clip_line, self.abs_tol)

    def clip_line(self, start: Vec2, end: Vec2) -> Sequence[tuple[Vec2, Vec2]]:
        """Returns the clipped line."""
        result = self._line_clipper.clip_line(start, end)
        if len(result) == 2:
            return (result,)  # type: ignore
        return tuple()

    def is_inside(self, point: Vec2) -> bool:
        """Returns ``True`` if `point` is inside the clipping rectangle."""
        return self._bbox.inside(point)

    def has_intersection(self, other: BoundingBox2d) -> bool:
        """Returns ``True`` if `other` bounding box intersects the clipping rectangle."""
        return self._bbox.has_intersection(other)


class ConcaveClippingPolygon2d:
    """The clipping path is an arbitrary concave 2D polygon."""

    def __init__(self, vertices: Iterable[Vec2], abs_tol=TOLERANCE):
        self.abs_tol = abs_tol
        clip = list(vertices)
        if len(clip) > 1:
            if clip[0].isclose(clip[-1], abs_tol=self.abs_tol):
                clip.pop()
        if len(clip) < 3:
            raise ValueError("more than 3 vertices as clipping polygon required")
        # open polygon; clockwise or counter-clockwise oriented vertices
        self._clipping_polygon = clip
        self._bbox = BoundingBox2d(clip)

    def is_inside(self, point: Vec2) -> bool:
        """Returns ``True`` if `point` is inside the clipping polygon."""
        if not self._bbox.inside(point):
            return False
        return (
            is_point_in_polygon_2d(point, self._clipping_polygon, abs_tol=self.abs_tol)
            >= 0
        )

    def clip_line(self, start: Vec2, end: Vec2) -> Sequence[tuple[Vec2, Vec2]]:
        """Returns the clipped line."""
        abs_tol = self.abs_tol
        line = (start, end)
        if not self._bbox.has_overlap(BoundingBox2d(line)):
            return tuple()

        intersections = polygon_line_intersections_2d(self._clipping_polygon, line)
        start_is_inside = is_point_in_polygon_2d(start, self._clipping_polygon) >= 0
        if len(intersections) == 0:
            if start_is_inside:
                return (line,)
            return tuple()
        end_is_inside = (
            is_point_in_polygon_2d(end, self._clipping_polygon, abs_tol=abs_tol) >= 0
        )
        if end_is_inside and not intersections[-1].isclose(end, abs_tol=abs_tol):
            # last inside-segment ends at end
            intersections.append(end)
        if start_is_inside and not intersections[0].isclose(start, abs_tol=abs_tol):
            # first inside-segment begins at start
            intersections.insert(0, start)

        # REMOVE duplicate intersection points at the beginning and the end -
        # these are caused by clipping at the connection point of two edges.
        # KEEP duplicate intersection points in between - these are caused by the
        # coincident edges of inverted clipping polygons.  These intersections points
        # are required for the inside/outside rule to work properly!
        if len(intersections) > 1 and intersections[0].isclose(
            intersections[1], abs_tol=abs_tol
        ):
            intersections.pop(0)
        if len(intersections) > 1 and intersections[-1].isclose(
            intersections[-2], abs_tol=abs_tol
        ):
            intersections.pop()

        if has_collinear_edge(self._clipping_polygon, start, end):
            # slow detection: doesn't work with inside/outside rule!
            # test if mid-point of intersection-segment is inside the polygon.
            # intersection-segment collinear with a polygon edge is inside!
            segments: list[tuple[Vec2, Vec2]] = []
            for a, b in pairwise(intersections):
                if a.isclose(b, abs_tol=abs_tol):  # ignore zero-length segments
                    continue
                if (
                    is_point_in_polygon_2d(
                        a.lerp(b), self._clipping_polygon, abs_tol=abs_tol
                    )
                    >= 0
                ):
                    segments.append((a, b))
            return segments

        # inside/outside rule
        # intersection segments:
        # (0, 1) outside (2, 3) outside (4, 5) ...
        return list(take2(intersections))

    def clip_polyline(self, polyline: Sequence[Vec2]) -> Sequence[Sequence[Vec2]]:
        """Returns the parts of the clipped polyline."""
        abs_tol = self.abs_tol
        segments: list[list[Vec2]] = []
        for start, end in pairwise(polyline):
            for a, b in self.clip_line(start, end):
                if segments:
                    last_seg = segments[-1]
                    if last_seg[-1].isclose(a, abs_tol=abs_tol):
                        last_seg.append(b)
                        continue
                segments.append([a, b])
        return segments

    def clip_polygon(self, polygon: Sequence[Vec2]) -> Sequence[Sequence[Vec2]]:
        """Returns the parts of the clipped polygon. A polygon is a closed polyline."""
        vertices = list(polygon)
        abs_tol = self.abs_tol
        if len(vertices) > 1:
            if vertices[0].isclose(vertices[-1], abs_tol=abs_tol):
                vertices.pop()
        if len(vertices) < 3:
            return tuple()
        polygon_box = BoundingBox2d(vertices)
        if not self._bbox.has_intersection(polygon_box):
            return tuple()  # polygons do not overlap
        result = clip_arbitrary_polygons(self._clipping_polygon, vertices)
        if len(result) == 0:
            is_outside = any(
                is_point_in_polygon_2d(v, self._clipping_polygon, abs_tol=abs_tol) < 0
                for v in vertices
            )
            if is_outside:
                return tuple()
            return (vertices,)
            # return (self._clipping_polygon.copy(),)
        return result


def clip_arbitrary_polygons(
    clipper: list[Vec2], subject: list[Vec2]
) -> Sequence[Sequence[Vec2]]:
    """Returns the parts of the clipped subject. Both polygons can be concave

    Args:
        clipper: clipping window closed polygon
        subject: closed polygon to clip

    """
    # Caching of gh_clipper is not possible, because both GHPolygons get modified!
    gh_clipper = GHPolygon.from_vec2(clipper)
    gh_subject = GHPolygon.from_vec2(subject)
    return gh_clipper.intersection(gh_subject)


def has_collinear_edge(polygon: list[Vec2], start: Vec2, end: Vec2) -> bool:
    """Returns ``True`` if `polygon` has any collinear edge to line `start->end`."""
    a = polygon[-1]
    rel_a = point_to_line_relation(a, start, end)
    for b in polygon:
        rel_b = point_to_line_relation(b, start, end)
        if rel_a == 0 and rel_b == 0:
            return True
        a = b
        rel_a = rel_b
    return False


def polygon_line_intersections_2d(
    polygon: list[Vec2], line: tuple[Vec2, Vec2], abs_tol: float = TOLERANCE
) -> list[Vec2]:
    """Returns all intersections of polygon with line.
    All intersections points are ordered from start to end of line.
    Start and end points are not included if not explicit intersection points.

    .. Note::

        Returns duplicate intersections points when the line intersects at
        the connection point of two polygon edges!

    """
    intersection_points: list[Vec2] = []
    start, end = line
    size = len(polygon)
    for index in range(size):
        a = polygon[index - 1]
        b = polygon[index]
        ip = intersection_line_line_2d((a, b), line, virtual=False, abs_tol=abs_tol)
        if ip is None:
            continue
        # Note: do not remove duplicate vertices, because inverted clipping polygons
        # have coincident clipping edges inside the clipping polygon! #1101
        if ip.isclose(a, abs_tol=abs_tol):
            a_prev = polygon[index - 2]
            rel_prev = point_to_line_relation(a_prev, start, end, abs_tol=abs_tol)
            rel_next = point_to_line_relation(b, start, end, abs_tol=abs_tol)
            if rel_prev == rel_next:
                continue
        # edge case: line intersects "exact" in point b
        elif ip.isclose(b, abs_tol=abs_tol):
            b_next = polygon[(index + 1) % size]
            rel_prev = point_to_line_relation(a, start, end, abs_tol=abs_tol)
            rel_next = point_to_line_relation(b_next, start, end, abs_tol=abs_tol)
            if rel_prev == rel_next:
                continue
        intersection_points.append(ip)

    intersection_points.sort(key=lambda ip: ip.distance(start))
    return intersection_points


class InvertedClippingPolygon2d(ConcaveClippingPolygon2d):
    """This class represents an inverted clipping path.  Everything between the inner
    polygon and the outer extents is considered as inside.  The inner clipping path is
    an arbitrary 2D polygon.

    .. Important::

        The `outer_bounds` must be larger than the content to clip to work correctly.

    """

    def __init__(
        self,
        inner_polygon: Iterable[Vec2],
        outer_bounds: BoundingBox2d,
        abs_tol=TOLERANCE,
    ):
        self.abs_tol = abs_tol
        clip = list(inner_polygon)
        if len(clip) > 1:
            if not clip[0].isclose(clip[-1], abs_tol=abs_tol):  # close inner_polygon
                clip.append(clip[0])
        if len(clip) < 4:
            raise ValueError("more than 3 vertices as clipping polygon required")
        # requirements for inner_polygon:
        # arbitrary polygon (convex or concave)
        # closed polygon (first vertex == last vertex)
        # clockwise or counter-clockwise oriented vertices
        self._clipping_polygon = make_inverted_clipping_polygon(
            clip, outer_bounds, abs_tol
        )
        self._bbox = outer_bounds


def make_inverted_clipping_polygon(
    inner_polygon: list[Vec2], outer_bounds: BoundingBox2d, abs_tol=TOLERANCE
) -> list[Vec2]:
    """Creates a closed inverted clipping polygon by connecting the inner polygon with
    the surrounding rectangle at their closest vertices.
    """
    assert outer_bounds.has_data is True
    inner_polygon = inner_polygon.copy()
    if inner_polygon[0].isclose(inner_polygon[-1], abs_tol=abs_tol):
        inner_polygon.pop()
    assert len(inner_polygon) > 2
    outer_rect = list(outer_bounds.rect_vertices())  # counter-clockwise
    outer_rect.reverse()  # clockwise
    ci, co = find_closest_vertices(inner_polygon, outer_rect)
    result = inner_polygon[ci:]
    result.extend(inner_polygon[: ci + 1])
    result.extend(outer_rect[co:])
    result.extend(outer_rect[: co + 1])
    result.append(result[0])
    return result


def find_closest_vertices(
    vertices0: list[Vec2], vertices1: list[Vec2]
) -> tuple[int, int]:
    """Returns the indices of the closest vertices of both lists."""
    min_dist = math.inf
    result: tuple[int, int] = 0, 0
    for i0, v0 in enumerate(vertices0):
        for i1, v1 in enumerate(vertices1):
            distance = v0.distance(v1)
            if distance < min_dist:
                min_dist = distance
                result = i0, i1
    return result


# Based on the paper "Efficient Clipping of Arbitrary Polygons" by
# Günther Greiner and Kai Hormann,
# ACM Transactions on Graphics 1998;17(2):71-83
# Available at: http://www.inf.usi.ch/hormann/papers/Greiner.1998.ECO.pdf


class _Node:
    def __init__(
        self,
        vtx: Vec2,
        alpha: float = 0.0,
        intersect=False,
        entry=True,
        checked=False,
    ):
        self.vtx = vtx

        # Reference to the next vertex of the polygon
        self.next: _Node = None  # type: ignore

        # Reference to the previous vertex of the polygon
        self.prev: _Node = None  # type: ignore

        # Reference to the corresponding intersection vertex in the other polygon
        self.neighbor: _Node = None  # type: ignore

        # True if intersection is an entry point, False if exit
        self.entry: bool = entry

        # Intersection point's relative distance from previous vertex
        self.alpha: float = alpha

        # True if vertex is an intersection
        self.intersect: bool = intersect

        # True if the vertex has been checked (last phase)
        self.checked: bool = checked

    def set_checked(self):
        self.checked = True
        if self.neighbor and not self.neighbor.checked:
            self.neighbor.set_checked()


class IntersectionError(Exception):
    pass


class GHPolygon:
    first: _Node = None  # type: ignore
    max_x: float = 1e6

    def add(self, node: _Node):
        """Add a polygon vertex node."""

        self.max_x = max(self.max_x, node.vtx.x)
        if self.first is None:
            self.first = node
            self.first.next = node
            self.first.prev = node
        else:  # insert as last node
            first = self.first
            last = first.prev
            first.prev = node
            node.next = first
            node.prev = last
            last.next = node

    @staticmethod
    def build(vertices: Iterable[UVec]) -> GHPolygon:
        """Build a new GHPolygon from an iterable of vertices."""
        return GHPolygon.from_vec2(Vec2.list(vertices))

    @staticmethod
    def from_vec2(vertices: Sequence[Vec2]) -> GHPolygon:
        """Build a new GHPolygon from an iterable of vertices."""
        polygon = GHPolygon()
        for v in vertices:
            polygon.add(_Node(v))
        return polygon

    @staticmethod
    def insert(vertex: _Node, start: _Node, end: _Node):
        """Insert and sort an intersection node.

        This function inserts an intersection node between two other
        start- and end node of an edge. The start and end node cannot be
        an intersection node (that is, they must be actual vertex nodes of
        the original polygon). If there are multiple intersection nodes
        between the start- and end node, then the new node is inserted
        based on its alpha value.
        """
        curr = start
        while curr != end and curr.alpha < vertex.alpha:
            curr = curr.next

        vertex.next = curr
        prev = curr.prev
        vertex.prev = prev
        prev.next = vertex
        curr.prev = vertex

    def __iter__(self) -> Iterator[_Node]:
        assert self.first is not None
        s = self.first
        while True:
            yield s
            s = s.next
            if s is self.first:
                return

    @property
    def first_intersect(self) -> Optional[_Node]:
        for v in self:
            if v.intersect and not v.checked:
                return v
        return None

    @property
    def points(self) -> list[Vec2]:
        points = [v.vtx for v in self]
        if not points[0].isclose(points[-1]):
            points.append(points[0])
        return points

    def unprocessed(self):
        for v in self:
            if v.intersect and not v.checked:
                return True
        return False

    def union(self, clip: GHPolygon) -> list[list[Vec2]]:
        return self.clip(clip, False, False)

    def intersection(self, clip: GHPolygon) -> list[list[Vec2]]:
        return self.clip(clip, True, True)

    def difference(self, clip: GHPolygon) -> list[list[Vec2]]:
        return self.clip(clip, False, True)

    # pylint: disable=too-many-branches
    def clip(self, clip: GHPolygon, s_entry, c_entry) -> list[list[Vec2]]:
        """Clip this polygon using another one as a clipper.

        This is where the algorithm is executed. It allows you to make
        a UNION, INTERSECT or DIFFERENCE operation between two polygons.

        Given two polygons A, B the following operations may be performed:

        A|B ... A OR B  (Union of A and B)
        A&B ... A AND B (Intersection of A and B)
        A\\B ... A - B
        B\\A ... B - A

        The entry records store the direction the algorithm should take when
        it arrives at that entry point in an intersection. Depending on the
        operation requested, the direction is set as follows for entry points
        (f=forward, b=backward; exit points are always set to the opposite):

              Entry
              A   B
              -----
        A|B   b   b
        A&B   f   f
        A\\B  b   f
        B\\A  f   b

        f = True, b = False when stored in the entry record
        """
        # Phase 1: Find intersections
        for subject_vertex in self:
            if not subject_vertex.intersect:
                for clipper_vertex in clip:
                    if not clipper_vertex.intersect:
                        ip, us, uc = line_intersection(
                            subject_vertex.vtx,
                            next_vertex_node(subject_vertex.next).vtx,
                            clipper_vertex.vtx,
                            next_vertex_node(clipper_vertex.next).vtx,
                        )
                        if ip is None:
                            continue
                        subject_node = _Node(ip, us, intersect=True, entry=False)
                        clipper_node = _Node(ip, uc, intersect=True, entry=False)
                        subject_node.neighbor = clipper_node
                        clipper_node.neighbor = subject_node

                        self.insert(
                            subject_node,
                            subject_vertex,
                            next_vertex_node(subject_vertex.next),
                        )
                        clip.insert(
                            clipper_node,
                            clipper_vertex,
                            next_vertex_node(clipper_vertex.next),
                        )

        # Phase 2: Identify entry/exit points
        s_entry ^= is_inside_polygon(self.first.vtx, clip)
        for subject_vertex in self:
            if subject_vertex.intersect:
                subject_vertex.entry = s_entry
                s_entry = not s_entry

        c_entry ^= is_inside_polygon(clip.first.vtx, self)
        for clipper_vertex in clip:
            if clipper_vertex.intersect:
                clipper_vertex.entry = c_entry
                c_entry = not c_entry

        # Phase 3: Construct clipped polygons
        clipped_polygons: list[list[Vec2]] = []
        while self.unprocessed():
            current: _Node = self.first_intersect  # type: ignore
            clipped: list[Vec2] = [current.vtx]
            while True:
                current.set_checked()
                if current.entry:
                    while True:
                        current = current.next
                        clipped.append(current.vtx)
                        if current.intersect:
                            break
                else:
                    while True:
                        current = current.prev
                        clipped.append(current.vtx)
                        if current.intersect:
                            break

                current = current.neighbor
                if current.checked:
                    break
            clipped_polygons.append(clipped)
        return clipped_polygons


def next_vertex_node(v: _Node) -> _Node:
    """Return the next non-intersecting vertex after the one specified."""
    c = v
    while c.intersect:
        c = c.next
    return c


def is_inside_polygon(vertex: Vec2, polygon: GHPolygon) -> bool:
    """Returns ``True`` if  `vertex` is inside `polygon`."""
    # Possible issue: are points on the boundary inside or outside the polygon?
    #  this version: inside
    return is_point_in_polygon_2d(vertex, polygon.points, abs_tol=TOLERANCE) >= 0


_ERROR = None, 0, 0


def line_intersection(
    s1: Vec2, s2: Vec2, c1: Vec2, c2: Vec2, tol: float = TOLERANCE
) -> tuple[Optional[Vec2], float, float]:
    """Returns the intersection point between two lines.

    This special implementation excludes the line end points as intersection
    points!

    Algorithm based on: http://paulbourke.net/geometry/lineline2d/
    """
    den = (c2.y - c1.y) * (s2.x - s1.x) - (c2.x - c1.x) * (s2.y - s1.y)
    if abs(den) < tol:
        return _ERROR
    us = ((c2.x - c1.x) * (s1.y - c1.y) - (c2.y - c1.y) * (s1.x - c1.x)) / den
    lwr = 0.0 + tol
    upr = 1.0 - tol
    # Line end points are excluded as intersection points:
    # us =~ 0.0; us =~ 1.0
    if not (lwr < us < upr):
        return _ERROR
    # uc =~ 0.0; uc =~ 1.0
    uc = ((s2.x - s1.x) * (s1.y - c1.y) - (s2.y - s1.y) * (s1.x - c1.x)) / den
    if lwr < uc < upr:
        return (
            Vec2(s1.x + us * (s2.x - s1.x), s1.y + us * (s2.y - s1.y)),
            us,
            uc,
        )
    return _ERROR


class BooleanOperation(enum.Enum):
    UNION = "union"
    DIFFERENCE = "difference"
    INTERSECTION = "intersection"


def greiner_hormann_intersection(
    p1: Iterable[UVec], p2: Iterable[UVec]
) -> list[list[Vec2]]:
    """Returns the INTERSECTION of polygon `p1` &  polygon `p2`.
    This algorithm works only for polygons with real intersection points
    and line end points on face edges are not considered as such intersection
    points!

    """
    return greiner_hormann(p1, p2, BooleanOperation.INTERSECTION)


def greiner_hormann_difference(
    p1: Iterable[UVec], p2: Iterable[UVec]
) -> list[list[Vec2]]:
    """Returns the DIFFERENCE of polygon `p1` - polygon `p2`.
    This algorithm works only for polygons with real intersection points
    and line end points on face edges are not considered as such intersection
    points!

    """
    return greiner_hormann(p1, p2, BooleanOperation.DIFFERENCE)


def greiner_hormann_union(p1: Iterable[UVec], p2: Iterable[UVec]) -> list[list[Vec2]]:
    """Returns the UNION of polygon `p1` | polygon `p2`.
    This algorithm works only for polygons with real intersection points
    and line end points on face edges are not considered as such intersection
    points!

    """
    return greiner_hormann(p1, p2, BooleanOperation.UNION)


def greiner_hormann(
    p1: Iterable[UVec], p2: Iterable[UVec], op: BooleanOperation
) -> list[list[Vec2]]:
    """Implements a 2d clipping function to perform 3 boolean operations:

    - UNION: p1 | p2 ... p1 OR p2
    - INTERSECTION: p1 & p2 ... p1 AND p2
    - DIFFERENCE: p1 \\ p2 ... p1 - p2

    Based on the paper "Efficient Clipping of Arbitrary Polygons" by
    Günther Greiner and Kai Hormann.
    This algorithm works only for polygons with real intersection points
    and line end points on face edges are not considered as such intersection
    points!

    """
    polygon1 = GHPolygon.build(p1)
    polygon2 = GHPolygon.build(p2)

    if op == BooleanOperation.UNION:
        return polygon1.union(polygon2)
    elif op == BooleanOperation.DIFFERENCE:
        return polygon1.difference(polygon2)
    elif op == BooleanOperation.INTERSECTION:
        return polygon1.intersection(polygon2)
    raise ValueError(f"unknown or unsupported boolean operation: {op}")


LEFT = 0x1
RIGHT = 0x2
BOTTOM = 0x4
TOP = 0x8


class CohenSutherlandLineClipping2d:
    """Cohen-Sutherland 2D line clipping algorithm, source:
    https://en.wikipedia.org/wiki/Cohen%E2%80%93Sutherland_algorithm

    Args:
        w_min: bottom-left corner of the clipping rectangle
        w_max: top-right corner of the clipping rectangle

    """

    __slots__ = ("x_min", "x_max", "y_min", "y_max")

    def __init__(self, w_min: Vec2, w_max: Vec2) -> None:
        self.x_min, self.y_min = w_min
        self.x_max, self.y_max = w_max

    def encode(self, x: float, y: float) -> int:
        code: int = 0
        if x < self.x_min:
            code |= LEFT
        elif x > self.x_max:
            code |= RIGHT
        if y < self.y_min:
            code |= BOTTOM
        elif y > self.y_max:
            code |= TOP
        return code

    def clip_line(self, p0: Vec2, p1: Vec2) -> Sequence[Vec2]:
        """Returns the clipped line part as tuple[Vec2, Vec2] or an empty tuple.

        Args:
            p0: start-point of the line to clip
            p1: end-point of the line to clip

        """
        x0, y0 = p0
        x1, y1 = p1
        code0 = self.encode(x0, y0)
        code1 = self.encode(x1, y1)
        x = x0
        y = y0
        while True:
            if not code0 | code1:  # ACCEPT
                # bitwise OR is 0: both points inside window; trivially accept and
                # exit loop:
                return Vec2(x0, y0), Vec2(x1, y1)
            if code0 & code1:  # REJECT
                # bitwise AND is not 0: both points share an outside zone (LEFT,
                # RIGHT, TOP, or BOTTOM), so both must be outside window;
                # exit loop
                return tuple()

            # failed both tests, so calculate the line segment to clip
            # from an outside point to an intersection with clip edge
            # At least one endpoint is outside the clip rectangle; pick it
            code = code1 if code1 > code0 else code0

            # Now find the intersection point;
            # use formulas:
            # slope = (y1 - y0) / (x1 - x0)
            # x = x0 + (1 / slope) * (ym - y0), where ym is y_min or y_max
            # y = y0 + slope * (xm - x0), where xm is x_min or x_max
            # No need to worry about divide-by-zero because, in each case, the
            # code bit being tested guarantees the denominator is non-zero
            if code & TOP:  # point is above the clip window
                x = x0 + (x1 - x0) * (self.y_max - y0) / (y1 - y0)
                y = self.y_max
            elif code & BOTTOM:  # point is below the clip window
                x = x0 + (x1 - x0) * (self.y_min - y0) / (y1 - y0)
                y = self.y_min
            elif code & RIGHT:  # point is to the right of clip window
                y = y0 + (y1 - y0) * (self.x_max - x0) / (x1 - x0)
                x = self.x_max
            elif code & LEFT:  # point is to the left of clip window
                y = y0 + (y1 - y0) * (self.x_min - x0) / (x1 - x0)
                x = self.x_min

            if code == code0:
                x0 = x
                y0 = y
                code0 = self.encode(x0, y0)
            else:
                x1 = x
                y1 = y
                code1 = self.encode(x1, y1)



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import io
import ezdxf
from ezdxf.math import Vec3
from ezdxf import bbox
from ezdxf.render import MeshBuilder


@pytest.fixture(scope="module")
def triangle_mesh():
    doc = ezdxf.read(io.StringIO(POLYMESH))
    msp = doc.modelspace()
    return msp[0]


@pytest.fixture
def virtual_3dface(triangle_mesh):
    virtual_entities = list(triangle_mesh.virtual_entities())
    return virtual_entities[0]


def test_is_poly_face_mesh(triangle_mesh):
    assert triangle_mesh.is_poly_face_mesh


def test_virtual_3dface_has_only_3_vertices(virtual_3dface):
    vertices = virtual_3dface.wcs_vertices()
    assert len(vertices) == 3


def test_virtual_3dface_does_not_include_NULLVEC(virtual_3dface):
    box = bbox.extents([virtual_3dface])
    assert box.extmin > Vec3(1, 1, 0)


def test_mesh_builder(triangle_mesh):
    builder = MeshBuilder.from_polyface(triangle_mesh)
    assert len(builder.vertices) == 3
    faces = list(builder.faces_as_vertices())
    assert len(faces) == 1, "expected only one face"
    assert len(faces[0]) == 3, "expected a triangle face"


if __name__ == "__main__":
    pytest.main([__file__])

# DXF R12, only ENTITIES section
POLYMESH = """0
SECTION
2
ENTITIES
0
POLYLINE
8
0
66
1
10
0.0
20
0.0
30
0.0
70
64
71
3
72
1
0
VERTEX
8
0
10
59416.31455091119
20
21790.91531877725
30
0.0
70
192
0
VERTEX
8
0
10
59361.25307228302
20
21813.84830555742
30
0.0
70
192
0
VERTEX
8
0
10
59380.16605928067
20
21806.95449032696
30
0.0
70
192
0
VERTEX
8
0
10
0.0
20
0.0
30
0.0
70
128
71
1
72
2
73
3
0
SEQEND
8
0
0
ENDSEC
0
EOF
"""



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities import DXFClass, is_dxf_object, is_graphic_entity
from ezdxf.lldxf.const import DXF12, DXF2000, DXF2004
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.tagwriter import TagCollector

HELIXCLS = """  0
CLASS
1
HELIX
2
AcDbHelix
3
ObjectDBX Classes
90
4095
91
0
280
0
281
1
"""


@pytest.fixture
def entity():
    return DXFClass.from_text(HELIXCLS)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "CLASS" in ENTITY_CLASSES


def test_default_init():
    dxfclass = DXFClass()
    assert dxfclass.dxftype() == "CLASS"
    assert dxfclass.dxf.handle is None
    assert dxfclass.dxf.owner is None


def test_default_new():
    entity = DXFClass.new(
        dxfattribs={
            "name": "HELIX",
            "cpp_class_name": "AcDbHelix",
            "flags": 4095,
            "app_name": "ObjectDBX Classes",
            "is_an_entity": 1,
        }
    )
    assert entity.dxf.name == "HELIX"
    assert entity.dxf.cpp_class_name == "AcDbHelix"
    assert entity.dxf.app_name == "ObjectDBX Classes"
    assert entity.dxf.flags == 4095
    assert entity.dxf.instance_count == 0
    assert entity.dxf.was_a_proxy == 0
    assert entity.dxf.is_an_entity == 1


def test_load_from_text(entity):
    assert entity.dxf.name == "HELIX"
    assert entity.dxf.cpp_class_name == "AcDbHelix"
    assert entity.dxf.app_name == "ObjectDBX Classes"
    assert entity.dxf.flags == 4095
    assert entity.dxf.instance_count == 0
    assert entity.dxf.was_a_proxy == 0
    assert entity.dxf.is_an_entity == 1


def test_write_dxf_2000(entity):
    expected = Tags.from_text(HELIXCLS)
    expected.remove_tags((91,))
    collector = TagCollector(dxfversion=DXF2000)
    entity.export_dxf(collector)
    assert collector.tags == expected


def test_write_dxf_2004(entity):
    expected = Tags.from_text(HELIXCLS)
    collector = TagCollector(dxfversion=DXF2004)
    entity.export_dxf(collector)
    assert collector.tags == expected


def test_write_dxf_r12(entity):
    collector = TagCollector(dxfversion=DXF12)
    entity.export_dxf(collector)
    assert len(collector.tags) == 0


def test_entity_type(entity):
    assert is_graphic_entity(entity) is False
    assert is_dxf_object(entity) is False



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
import enum
from xml.etree import ElementTree as ET

import ezdxf
from ezdxf.document import Drawing
from ezdxf import zoom, colors
from ezdxf.addons import xplayer
from ezdxf.addons.drawing import svg, layout, pymupdf, dxf
from ezdxf.addons.drawing.dxf import ColorMode

from .tokenizer import hpgl2_commands
from .plotter import Plotter
from .interpreter import Interpreter
from .backend import Recorder, placement_matrix, Player


DEBUG = False
ENTER_HPGL2_MODE = b"%1B"


class Hpgl2Error(Exception):
    """Base exception for the :mod:`hpgl2` add-on."""

    pass


class Hpgl2DataNotFound(Hpgl2Error):
    """No HPGL/2 data was found, maybe the "Enter HPGL/2 mode" escape sequence is missing."""

    pass


class EmptyDrawing(Hpgl2Error):
    """The HPGL/2 commands do not produce any content."""

    pass


class MergeControl(enum.IntEnum):
    NONE = 0  # print order
    LUMINANCE = 1  # sort filled polygons by luminance
    AUTO = 2  # guess best method


def to_dxf(
    b: bytes,
    *,
    rotation: int = 0,
    mirror_x: bool = False,
    mirror_y: bool = False,
    color_mode=ColorMode.RGB,
    merge_control: MergeControl = MergeControl.AUTO,
) -> Drawing:
    """
    Exports the HPGL/2 commands of the byte stream `b` as a DXF document.

    The page content is created at the origin of the modelspace and 1 drawing unit is 1
    plot unit (1 plu = 0.025mm) unless scaling values are provided.

    The content of HPGL files is intended to be plotted on white paper, therefore a white
    filling will be added as background in color mode :attr:`RGB`.

    All entities are assigned to a layer according to the pen number with the name scheme
    ``PEN_<###>``. In order to be able to process the file better, it is also possible to
    assign the :term:`ACI` color by layer by setting the argument `color_mode` to
    :attr:`ColorMode.ACI`, but then the RGB color is lost because the RGB color has
    always the higher priority over the :term:`ACI`.

    The first paperspace layout "Layout1" of the DXF document is set up to print the entire
    modelspace on one sheet, the size of the page is the size of the original plot file in
    millimeters.

    HPGL/2's merge control works at the pixel level and cannot be replicated by DXF,
    but to prevent fillings from obscuring text, the filled polygons are
    sorted by luminance - this can be forced or disabled by the argument `merge_control`,
    see also :class:`MergeControl` enum.

    Args:
        b: plot file content as bytes
        rotation: rotation angle of 0, 90, 180 or 270 degrees
        mirror_x: mirror in x-axis direction
        mirror_y:  mirror in y-axis direction
        color_mode: the color mode controls how color values are assigned to DXF entities,
            see :class:`ColorMode`
        merge_control: how to order filled polygons, see :class:`MergeControl`

    Returns: DXF document as instance of class :class:`~ezdxf.document.Drawing`

    """
    if rotation not in (0, 90, 180, 270):
        raise ValueError("rotation angle must be 0, 90, 180, or 270")

    # 1st pass records output of the plotting commands and detects the bounding box
    doc = ezdxf.new()
    try:
        player = record_plotter_output(b, merge_control)
    except Hpgl2Error:
        return doc

    bbox = player.bbox()
    m = placement_matrix(bbox, -1 if mirror_x else 1, -1 if mirror_y else 1, rotation)
    player.transform(m)
    bbox = player.bbox()
    msp = doc.modelspace()
    dxf_backend = dxf.DXFBackend(msp, color_mode=color_mode)
    bg_color = colors.RGB(255, 255, 255)
    if color_mode == ColorMode.RGB:
        doc.layers.add("BACKGROUND")
        bg = dxf.add_background(msp, bbox, color=bg_color)
        bg.dxf.layer = "BACKGROUND"

    # 2nd pass replays the plotting commands to plot the DXF
    # exports the HPGL/2 content in plot units (plu) as modelspace:
    # 1 plu = 0.025mm or 40 plu == 1mm
    xplayer.hpgl2_to_drawing(player, dxf_backend, bg_color=bg_color.to_hex())
    del player

    if bbox.has_data:  # non-empty page
        zoom.window(msp, bbox.extmin, bbox.extmax)
        dxf.update_extents(doc, bbox)
        # paperspace is set up in mm:
        dxf.setup_paperspace(doc, bbox)
    return doc


def to_svg(
    b: bytes,
    *,
    rotation: int = 0,
    mirror_x: bool = False,
    mirror_y: bool = False,
    merge_control=MergeControl.AUTO,
) -> str:
    """
    Exports the HPGL/2 commands of the byte stream `b` as SVG string.

    The plot units are mapped 1:1 to ``viewBox`` units and the size of image is the size
    of the original plot file in millimeters.

    HPGL/2's merge control works at the pixel level and cannot be replicated by the
    backend, but to prevent fillings from obscuring text, the filled polygons are
    sorted by luminance - this can be forced or disabled by the argument `merge_control`,
    see also :class:`MergeControl` enum.

    Args:
        b: plot file content as bytes
        rotation: rotation angle of 0, 90, 180 or 270 degrees
        mirror_x: mirror in x-axis direction
        mirror_y:  mirror in y-axis direction
        merge_control: how to order filled polygons, see :class:`MergeControl`

    Returns: SVG content as ``str``

    """
    if rotation not in (0, 90, 180, 270):
        raise ValueError("rotation angle must be 0, 90, 180, or 270")
    # 1st pass records output of the plotting commands and detects the bounding box
    try:
        player = record_plotter_output(b, merge_control)
    except Hpgl2Error:
        return ""

    # transform content for the SVGBackend of the drawing add-on
    bbox = player.bbox()
    size = bbox.size

    # HPGL/2 uses (integer) plot units, 1 plu = 0.025mm or 1mm = 40 plu
    width_in_mm = size.x / 40
    height_in_mm = size.y / 40

    if rotation in (0, 180):
        page = layout.Page(width_in_mm, height_in_mm)
    else:
        page = layout.Page(height_in_mm, width_in_mm)
        # adjust rotation for y-axis mirroring
        rotation += 180

    # The SVGBackend expects coordinates in the range [0, output_coordinate_space]
    max_plot_units = round(max(size.x, size.y))
    settings = layout.Settings(output_coordinate_space=max_plot_units)
    m = layout.placement_matrix(
        bbox,
        sx=-1 if mirror_x else 1,
        sy=1 if mirror_y else -1,  # inverted y-axis
        rotation=rotation,
        page=page,
        output_coordinate_space=settings.output_coordinate_space,
    )
    player.transform(m)

    # 2nd pass replays the plotting commands on the SVGBackend of the drawing add-on
    svg_backend = svg.SVGRenderBackend(page, settings)
    xplayer.hpgl2_to_drawing(player, svg_backend, bg_color="#ffffff")
    del player
    xml = svg_backend.get_xml_root_element()
    return ET.tostring(xml, encoding="unicode", xml_declaration=True)


def to_pdf(
    b: bytes,
    *,
    rotation: int = 0,
    mirror_x: bool = False,
    mirror_y: bool = False,
    merge_control=MergeControl.AUTO,
) -> bytes:
    """
    Exports the HPGL/2 commands of the byte stream `b` as PDF data.

    The plot units (1 plu = 0.025mm) are converted to PDF units (1/72 inch) so the image
    has the size of the original plot file.

    HPGL/2's merge control works at the pixel level and cannot be replicated by the
    backend, but to prevent fillings from obscuring text, the filled polygons are
    sorted by luminance - this can be forced or disabled by the argument `merge_control`,
    see also :class:`MergeControl` enum.

    Python module PyMuPDF is required: https://pypi.org/project/PyMuPDF/

    Args:
        b: plot file content as bytes
        rotation: rotation angle of 0, 90, 180 or 270 degrees
        mirror_x: mirror in x-axis direction
        mirror_y:  mirror in y-axis direction
        merge_control: how to order filled polygons, see :class:`MergeControl`

    Returns: PDF content as ``bytes``

    """
    return _pymupdf(
        b,
        rotation=rotation,
        mirror_x=mirror_x,
        mirror_y=mirror_y,
        merge_control=merge_control,
        fmt="pdf",
    )


def to_pixmap(
    b: bytes,
    *,
    rotation: int = 0,
    mirror_x: bool = False,
    mirror_y: bool = False,
    merge_control=MergeControl.AUTO,
    fmt: str = "png",
    dpi: int = 96,
) -> bytes:
    """
    Exports the HPGL/2 commands of the byte stream `b` as pixel image.

    Supported image formats:

        === =========================
        png Portable Network Graphics
        ppm Portable Pixmap
        pbm Portable Bitmap
        === =========================

    The plot units (1 plu = 0.025mm) are converted to dot per inch (dpi) so the image
    has the size of the original plot file.

    HPGL/2's merge control works at the pixel level and cannot be replicated by the
    backend, but to prevent fillings from obscuring text, the filled polygons are
    sorted by luminance - this can be forced or disabled by the argument `merge_control`,
    see also :class:`MergeControl` enum.

    Python module PyMuPDF is required: https://pypi.org/project/PyMuPDF/

    Args:
        b: plot file content as bytes
        rotation: rotation angle of 0, 90, 180 or 270 degrees
        mirror_x: mirror in x-axis direction
        mirror_y:  mirror in y-axis direction
        merge_control: how to order filled polygons, see :class:`MergeControl`
        fmt: image format
        dpi: output resolution in dots per inch

    Returns: image content as ``bytes``

    """
    fmt = fmt.lower()
    if fmt not in pymupdf.SUPPORTED_IMAGE_FORMATS:
        raise ValueError(f"image format '{fmt}' not supported")
    return _pymupdf(
        b,
        rotation=rotation,
        mirror_x=mirror_x,
        mirror_y=mirror_y,
        merge_control=merge_control,
        fmt=fmt,
        dpi=dpi,
    )


def _pymupdf(
    b: bytes,
    *,
    rotation: int = 0,
    mirror_x: bool = False,
    mirror_y: bool = False,
    merge_control=MergeControl.AUTO,
    fmt: str = "pdf",
    dpi=96,
) -> bytes:
    if not pymupdf.is_pymupdf_installed:
        print("Python module PyMuPDF is required: https://pypi.org/project/PyMuPDF/")
        return b""

    if rotation not in (0, 90, 180, 270):
        raise ValueError("rotation angle must be 0, 90, 180, or 270")
    # 1st pass records output of the plotting commands and detects the bounding box
    try:
        player = record_plotter_output(b, merge_control)
    except Hpgl2Error:
        return b""

    # transform content for the SVGBackend of the drawing add-on
    bbox = player.bbox()
    size = bbox.size

    # HPGL/2 uses (integer) plot units, 1 plu = 0.025mm or 1mm = 40 plu
    width_in_mm = size.x / 40
    height_in_mm = size.y / 40

    if rotation in (0, 180):
        page = layout.Page(width_in_mm, height_in_mm)
    else:
        page = layout.Page(height_in_mm, width_in_mm)
        # adjust rotation for y-axis mirroring
        rotation += 180

    # The PDFBackend expects coordinates as pt = 1/72 inch; 1016 plu = 1 inch
    max_plot_units = max(size.x, size.y) / 1016 * 72
    settings = layout.Settings(output_coordinate_space=max_plot_units)
    m = layout.placement_matrix(
        bbox,
        sx=-1 if mirror_x else 1,
        sy=-1 if mirror_y else 1,
        rotation=rotation,
        page=page,
        output_coordinate_space=settings.output_coordinate_space,
    )
    player.transform(m)

    # 2nd pass replays the plotting commands on the PyMuPdfBackend of the drawing add-on
    pymupdf_backend = pymupdf.PyMuPdfBackend()
    xplayer.hpgl2_to_drawing(player, pymupdf_backend, bg_color="#ffffff")
    del player
    if fmt == "pdf":
        return pymupdf_backend.get_pdf_bytes(page, settings=settings)
    else:
        return pymupdf_backend.get_pixmap_bytes(
            page, fmt=fmt, settings=settings, dpi=dpi
        )


def print_interpreter_log(interpreter: Interpreter) -> None:
    print("HPGL/2 interpreter log:")
    print(f"unsupported commands: {interpreter.not_implemented_commands}")
    if interpreter.errors:
        print("parsing errors:")
        for err in interpreter.errors:
            print(err)


def record_plotter_output(
    b: bytes,
    merge_control: MergeControl,
) -> Player:
    commands = hpgl2_commands(b)
    if len(commands) == 0:
        print("HPGL2 data not found.")
        raise Hpgl2DataNotFound

    recorder = Recorder()
    plotter = Plotter(recorder)
    interpreter = Interpreter(plotter)
    interpreter.run(commands)
    if DEBUG:
        print_interpreter_log(interpreter)
    player = recorder.player()
    bbox = player.bbox()
    if not bbox.has_data:
        raise EmptyDrawing

    if merge_control == MergeControl.AUTO:
        if plotter.has_merge_control:
            merge_control = MergeControl.LUMINANCE
    if merge_control == MergeControl.LUMINANCE:
        if DEBUG:
            print("merge control on: sorting filled polygons by luminance")
        player.sort_filled_paths()
    return player



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf import appsettings
from ezdxf.math import Vec3


def test_update_extents():
    doc = ezdxf.new()
    msp = doc.modelspace()
    msp.add_line((0, 0), (7, 8, 9))
    appsettings.update_extents(doc)

    assert msp.dxf.extmin.isclose((0, 0))
    assert msp.dxf.extmax.isclose((7, 8, 9))

    assert Vec3(0, 0).isclose(doc.header["$EXTMIN"])
    assert Vec3(7, 8, 9).isclose(doc.header["$EXTMAX"])


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2010-2022, Manfred Moitzi
# License: MIT License
from typing import Sequence, Union, Callable, Any, NamedTuple, Optional
from .types import DXFVertex, DXFTag, cast_tag_value


def SingleValue(value: Union[str, float], code: int = 1) -> DXFTag:
    return DXFTag(code, cast_tag_value(code, value))


def Point2D(value: Sequence[float]) -> DXFVertex:
    return DXFVertex(10, (value[0], value[1]))


def Point3D(value: Sequence[float]) -> DXFVertex:
    return DXFVertex(10, (value[0], value[1], value[2]))


class HeaderVarDef(NamedTuple):
    name: str
    code: int
    factory: Callable[[Any], Any]
    mindxf: str
    maxdxf: str
    priority: int
    default: Optional[Any] = None



# cython: language_level=3
# Copyright (c) 2020-2023, Manfred Moitzi
# License: MIT License
from .vector cimport Vec3

cdef class Matrix44:
    cdef double m[16]
    cdef Vec3 get_ux(self: Matrix44)
    cdef Vec3 get_uy(self: Matrix44)
    cdef Vec3 get_uz(self: Matrix44)

cdef inline swap(double *a, double *b):
    cdef double tmp = a[0]
    a[0] = b[0]
    b[0] = tmp


# Copyright (c) , Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Sequence
import pytest

import math
from ezdxf import edgesmith as es
from ezdxf import edgeminer as em
from ezdxf.entities import Circle, Arc, Ellipse, LWPolyline, Spline
from ezdxf.math import fit_points_to_cad_cv
from ezdxf.layouts import VirtualLayout


def test_circle_is_a_closed_entity():
    circle = Circle()
    circle.dxf.radius = 1

    assert es.is_closed_entity(circle) is True


def test_circle_of_radius_0_is_not_a_closed_entity():
    circle = Circle()
    circle.dxf.radius = 0

    assert es.is_closed_entity(circle) is False


@pytest.mark.parametrize("start,end", [(0, 180), (0, 0), (180, 180), (360, 360)])
def test_open_arc_is_not_a_closed_entity(start, end):
    arc = Arc()
    arc.dxf.radius = 1
    arc.dxf.start_angle = start
    arc.dxf.end_angle = end

    assert es.is_closed_entity(arc) is False


@pytest.mark.parametrize("start,end", [(0, 360), (360, 0), (180, -180)])
def test_closed_arc_is_a_closed_entity(start, end):
    arc = Arc()
    arc.dxf.radius = 1
    arc.dxf.start_angle = start
    arc.dxf.end_angle = end

    assert es.is_closed_entity(arc) is True


@pytest.mark.parametrize(
    "start,end", [(0, math.pi), (0, 0), (math.pi, math.pi), (math.tau, math.tau)]
)
def test_open_ellipse_is_not_a_closed_entity(start, end):
    ellipse = Ellipse()
    ellipse.dxf.major_axis = (1, 0)
    ellipse.dxf.ratio = 1
    ellipse.dxf.start_param = start
    ellipse.dxf.end_param = end

    assert es.is_closed_entity(ellipse) is False


@pytest.mark.parametrize(
    "start,end", [(0, math.tau), (math.tau, 0), (math.pi, -math.pi)]
)
def test_closed_ellipse_is_a_closed_entity(start, end):
    ellipse = Ellipse()
    ellipse.dxf.major_axis = (1, 0)
    ellipse.dxf.ratio = 1
    ellipse.dxf.start_param = start
    ellipse.dxf.end_param = end

    assert es.is_closed_entity(ellipse) is True


# Note: Ellipses with major_axis == (0, 0, 0) are not valid.
# They cannot be created by ezdxf and loading such ellipses raises an DXFStructureError.


def test_closed_lwpolyline_is_a_closed_entity():
    polyline = LWPolyline()
    polyline.closed = True
    polyline.append_points([(0, 0), (10, 0), (10, 10)])

    assert es.is_closed_entity(polyline) is True


def test_open_lwpolyline_is_not_a_closed_entity():
    polyline = LWPolyline()
    polyline.closed = False
    polyline.append_points([(0, 0), (10, 0), (10, 10)])

    assert es.is_closed_entity(polyline) is False


def test_explicit_closed_lwpolyline_is_a_closed_entity():
    polyline = LWPolyline()
    polyline.closed = False
    polyline.append_points([(0, 0), (10, 0), (10, 10), (0, 0)])

    assert es.is_closed_entity(polyline) is True


def test_closed_spline():
    ct = fit_points_to_cad_cv([(0, 0), (10, 0), (10, 10), (0, 0)])
    spline = Spline()
    spline.apply_construction_tool(ct)

    assert es.is_closed_entity(spline) is True


def test_open_spline():
    ct = fit_points_to_cad_cv([(0, 0), (10, 0), (10, 10), (0, 10)])
    spline = Spline()
    spline.apply_construction_tool(ct)

    assert es.is_closed_entity(spline) is False


def test_empty_spline():
    spline = Spline()

    assert es.is_closed_entity(spline) is False


class TestLineAndArcToPolyline:
    @pytest.fixture(scope="class")
    def edges(self) -> Sequence[em.Edge]:
        layout = VirtualLayout()
        layout.add_line((0, 0), (6, 0))
        layout.add_arc((6, 3), radius=3, start_angle=270, end_angle=90)
        layout.add_line((6, 6), (0, 6))
        layout.add_line((0, 6), (0, 0))
        return list(es.edges_from_entities_2d(layout))

    def test_create_lwpolyline(self, edges: Sequence[em.Edge]):
        def int_pnt(point):
            return tuple(round(v) for v in point)

        pl = es.lwpolyline_from_chain(edges)
        assert len(pl.lwpoints) == 5
        assert int_pnt(pl.lwpoints[1]) == (6, 0, 0, 0, 1)
        assert int_pnt(pl.lwpoints[2]) == (6, 6, 0, 0, 0)

    def test_create_polyline2d(self, edges: Sequence[em.Edge]):
        def int_pnt(vertex):
            dxf = vertex.dxf
            return round(dxf.location.x), round(dxf.location.y), round(dxf.bulge)

        pl = es.polyline2d_from_chain(edges)
        assert len(pl.vertices) == 5

        assert int_pnt(pl.vertices[1]) == (6, 0, 1)
        assert int_pnt(pl.vertices[2]) == (6, 6, 0)

    def test_create_polyline_path(self, edges: Sequence[em.Edge]):
        # HATCH boundary path: ezdxf.entities.PolylinePath
        def int_pnt(point):
            return tuple(round(v) for v in point)

        pl_path = es.polyline_path_from_chain(edges)
        assert len(pl_path.vertices) == 5
        assert int_pnt(pl_path.vertices[1]) == (6, 0, 1)
        assert int_pnt(pl_path.vertices[2]) == (6, 6, 0)

    def test_create_edge_path(self, edges: Sequence[em.Edge]):
        # HATCH boundary path: ezdxf.entities.EdgePath
        pass

    def test_create_path(self, edges: Sequence[em.Edge]):
        # ezdxf.path.Path
        path = es.path2d_from_chain(edges)
        commands = path.commands()
        assert len(commands) == 5
        assert path.start.isclose((0, 0))
        assert path.is_closed is True
        assert path.has_curves is True
        assert path.has_lines is True


class TestBoundingBox:
    def test_empty_sequence(self):
        bbox = es.bounding_box_2d([])
        assert bbox.has_data is False

    def test_one_edge(self):
        edge = em.make_edge((0, 0), (1, 1))
        bbox = es.bounding_box_2d([edge])
        assert bbox.extmin.isclose((0, 0))
        assert bbox.extmax.isclose((1, 1))

    def test_more_edges(self):
        edge1 = em.make_edge((0, 0), (3, 4))
        edge2 = em.make_edge((-1, -2), (0, 0))
        bbox = es.bounding_box_2d([edge1, edge2])
        assert bbox.extmin.isclose((-1, -2))
        assert bbox.extmax.isclose((3, 4))


class TestIsInsidePolygon:
    @pytest.fixture
    def edges(self):
        return [
            em.make_edge((0, 0), (1, 0)),
            em.make_edge((1, 0), (1, 1)),
            em.make_edge((1, 1), (0, 1)),
            em.make_edge((0, 1), (0, 0)),
        ]

    def test_is_inside_polygon(self, edges):
        assert es.is_inside_polygon_2d(edges, (0.5, 0.5)) is True

    def test_on_border_is_inside_polygon(self, edges):
        assert es.is_inside_polygon_2d(edges, (0.5, 1.0)) is True

    def test_is_outside_polygon(self, edges):
        assert es.is_inside_polygon_2d(edges, (0.5, 1.5)) is False

    def test_raise_exception_if_not_a_loop(self, edges):
        with pytest.raises(ValueError):
            es.is_inside_polygon_2d(edges[:-1], (0.5, 0.5))


class TestIntersectingEdges:
    @pytest.fixture
    def edges(self):
        return [
            em.make_edge((0, -1), (0, 1)),
            em.make_edge((1, -1), (1, 1)),
            em.make_edge((2, -1), (2, 1)),
            em.make_edge((3, -1), (3, 1)),
        ]

    def test_empty_sequence(self, edges):
        assert len(es.intersecting_edges_2d([], (-1, 0.5))) == 0

    def test_no_intersection(self, edges):
        assert len(es.intersecting_edges_2d(edges, (-2, 0), (-1, 0))) == 0

    def test_first_intersection(self, edges):
        result = es.intersecting_edges_2d(edges, (-2, 0), (0.5, 0))
        assert result[0].edge.id == edges[0].id
        assert math.isclose(result[0].distance, 2) is True

    def test_all_intersections(self, edges):
        result = es.intersecting_edges_2d(edges, (-2, 0), (4, 0))
        assert len(result) == len(edges)

    def test_arbitrary_p2_v1(self, edges):
        result = es.intersecting_edges_2d(edges, (-2, 0))
        assert len(result) == len(edges)

    def test_arbitrary_p2_v2(self, edges):
        result = es.intersecting_edges_2d(edges, (4, 0))
        assert len(result) == 0


if __name__ == "__main__":
    pytest.main([__file__])
